"""
Trame example that checks an input DICOM file patient tag against a patient name in the trame state.
"""

from pathlib import Path
from typing import Literal

from trame.app import TrameApp
from trame.ui.vuetify3 import SinglePageLayout

from trame.widgets import client, vuetify3, html

# 1 - lookup the path of your JS file that contains your preprocessing logic
DICOM_UTILS_SCRIPT = Path(__file__).with_name("dicom_utils.js")


class PatientNameValidator(TrameApp):
    def __init__(self, server=None):
        super().__init__(server)

        self.state.patient_name = "John Doe"

        # 2 - register your JS file and entrypoint function using trame_client.Handler module API
        client.register_external_script(
            name="validate_dicom_patient_name",
            script_file_path=DICOM_UTILS_SCRIPT,
            function_names=["validateDicomFilePatientName"],
        )

        self._build_ui()

    def _build_ui(self):
        with SinglePageLayout(self.server) as self.ui:
            self.ui.title.set_text("Client Handler Preprocessing Minimal Example")

            with (
                self.ui.content,
                vuetify3.VContainer(fluid=True),
                vuetify3.VRow(),
                vuetify3.VCol(cols=3),
                # 3. Instantiate the PreProcessor widget
                client.Handler(
                    function="validate_dicom_patient_name",
                    # 4. Pass additional inputs to your preprocessing logic
                    # those must match your function signature in the JS logic file
                    # this is a JS expression that will be passed as is to your function
                    inputs=("{ patient_name }",),
                    # 5. Register a callback to be run when your preprocessing is completed
                    completed=(
                        self.preprocessing_pipeline_completed,
                        "[$event.type, $event.outputs]",
                    ),
                ) as client_handler,
            ):
                with html.Div(style="display: flex; flex-direction: column;"):
                    vuetify3.VFileInput(
                        # 6. Use the provided input_data from the widget to trigger the preprocessing on it
                        # preprocessing.input_data is a provided slot scopped reactive value
                        # which is observed by the PreProcessor component
                        # it should be assigned with whatever data your want to run preprocessing on
                        v_model=client_handler.input_data,
                        hint="Input here a DICOM file",
                        label="DICOM file",
                        persistent_hint=True,
                    )

                    vuetify3.VTextField(
                        v_model="patient_name",
                        hint="Input here the patient name that you expect for your DICOM file",
                        label="Expected Patient Name",
                        persistent_hint=True,
                    )

                    vuetify3.VAlert(
                        v_if="alert_title",
                        text=("alert_text", None),
                        title=("alert_title", None),
                        type=("alert_type", None),
                    )

    def preprocessing_pipeline_completed(
        self, type: Literal["success", "failure"], output
    ) -> None:
        print("preprocessing_pipeline_completed")
        print(type)
        print(output)

        self.state.alert_type = type if type == "success" else "error"
        self.state.alert_title = self.state.alert_type

        if type == "failure":
            self.state.alert_text = (
                f"Expected patient {output[0]['expected']}, got {output[0]['got']}"
            )
        else:
            self.state.alert_text = "Patient name matched! We got the file."


def main(**kwargs):
    app = PatientNameValidator()
    app.server.start(**kwargs)


if __name__ == "__main__":
    main()



from .core import AbstractLayout
from ..widgets.html import Div

__all__ = [
    "DivLayout",
]


class DivLayout(AbstractLayout):
    def __init__(
        self,
        _server,
        template_name="main",
        connect_parent=True,
        style=None,
        classes=None,
        **kwargs,
    ):
        """
        Simple <div> element as a layout

        :param _server: Server to bound the layout to
        :param template_name: Name of the template (default: main)
        :param connect_parent: Control registration to parent context manager
        :param style: Forwarded to root Div element
        :param classes: Forwarded to root Div element
        :param **kwargs: Optional arguments pass to the <Div> element
        """
        super().__init__(
            _server,
            Div(
                trame_server=_server,
                connect_parent=connect_parent,
                style=style,
                classes=classes,
            ),
            template_name=template_name,
            **kwargs,
        )



from trame.app import get_server
from trame_client.utils.testing import enable_testing

server = get_server(client_type="vue3")
server.state.count = 1
server.state.trame__template_main = """
    <div>
        <div class="countValue" style="padding: 20px; background: red;">{{ count }}</div>
        <button class="plusButton" @click="count++">Add to count</button>
    </div>
"""

enable_testing(server, "count")
server.start()



import asyncio
import pytest

from trame_client.widgets.core import TrameComponent
from trame.decorators import change, controller, trigger
from trame.app import get_server

from trame_client.ui.html import DivLayout
from trame_client.widgets import html

CLIENT_TYPE = "vue3"


class TestComponent(TrameComponent):
    def __init__(self, server):
        self._steps = []
        super().__init__(server)

    def validate(self):
        assert self._steps == [
            "a changed to 1",
            "a changed to 2",
            "ctrl.b() # set",
            "ctrl.c() # add",
        ]

    @change("a")
    def on_a(self, a, **_):
        self._steps.append(f"a changed to {a}")

    @controller.set("b")
    def on_ctrl_b(self):
        self._steps.append("ctrl.b() # set")

    @controller.add("c")
    def on_ctrl_c(self):
        self._steps.append("ctrl.c() # add")

    @trigger("hello")
    def on_trigger(self): ...


class WidgetComponent(html.Div):
    def __init__(self, **kwargs):
        self._steps = []
        super().__init__(**kwargs)
        self.state.setdefault("a", 1)

        with self:
            html.Label("Hello World {{ a }}")

    def validate(self):
        assert self._steps == [
            "a changed to 1",
            "a changed to 2",
            "ctrl.b() # set",
            "ctrl.c() # add",
        ]

    @change("a")
    def on_a(self, a, **_):
        self._steps.append(f"a changed to {a}")

    @controller.set("b")
    def on_ctrl_b(self):
        self._steps.append("ctrl.b() # set")

    @controller.add("c")
    def on_ctrl_c(self):
        self._steps.append("ctrl.c() # add")

    @trigger("hello")
    def on_trigger(self): ...


@pytest.mark.asyncio
async def test_component():
    server = get_server("test_component", client_type=CLIENT_TYPE)
    server.state.a = 1

    component = TestComponent(server)

    server.state.ready()
    await asyncio.sleep(0.1)

    with server.state:
        server.state.a = 2

    component.ctrl.b()
    component.ctrl.c()

    assert server.trigger_name(component.on_trigger) == "hello"

    await asyncio.sleep(0.1)

    component.validate()


@pytest.mark.asyncio
async def test_widget_as_component():
    server = get_server("test_widget_as_component", client_type=CLIENT_TYPE)

    with DivLayout(server) as layout:
        WidgetComponent(ctx_name="test_comp")

    assert (
        layout.html
        == """<div >
<div >
<label >
Hello World {{ a }}
</label>
</div>
</div>"""
    )

    server.state.ready()
    await asyncio.sleep(0.1)

    with server.state:
        server.state.a = 2

    server.controller.b()
    server.controller.c()

    assert server.trigger_name(server.context.test_comp.on_trigger) == "hello"

    await asyncio.sleep(0.1)

    server.context.test_comp.validate()



import asyncio
import random
from trame.app import get_server

server = get_server()
state, ctrl = server.state, server.controller

state.count = 1
state.play = False


@state.change("count")
def count_change(count, **_):
    print(f"count={count}")


@ctrl.trigger("add")
def add_to_count():
    state.count += 1


@ctrl.trigger("subtract")
def subtract_to_count():
    state.count -= 1


@ctrl.trigger("toggle_play")
def toggle_play():
    state.play = not state.play


@ctrl.trigger("random")
def random_count():
    server.js_call("counter", "reset", random.random())


async def animate(**kwargs):
    while True:
        await asyncio.sleep(0.5)
        if state.play:
            with state:
                state.count += 1


ctrl.on_server_ready.add_task(animate)

server.start()



from pathlib import Path

USER_PROVIDED_SCRIPTS_DIR_PATH = Path(__file__).with_name("user_provided_scripts")
USER_PROVIDED_UMD_SCRIPTS_DIR_PATH = USER_PROVIDED_SCRIPTS_DIR_PATH / "umd"
USER_PROVIDED_ES_SCRIPTS_DIR_PATH = USER_PROVIDED_SCRIPTS_DIR_PATH / "es"
USER_PROVIDED_SCRIPTS_SERVE_URL_PREFIX = Path("__trame_client_external_scripts")


def setup_handler_module(server):
    serve_path = USER_PROVIDED_SCRIPTS_DIR_PATH.resolve()
    module_scripts_path = USER_PROVIDED_ES_SCRIPTS_DIR_PATH.resolve()
    umd_scripts_path = USER_PROVIDED_UMD_SCRIPTS_DIR_PATH.resolve()

    server.enable_module(
        {
            "serve": {
                USER_PROVIDED_SCRIPTS_SERVE_URL_PREFIX.as_posix(): str(serve_path)
            },
            "module_scripts": [
                (
                    USER_PROVIDED_SCRIPTS_SERVE_URL_PREFIX
                    / module_scripts_path.name
                    / module_file.name
                ).as_posix()
                for module_file in module_scripts_path.glob("*.js")
            ],
            "scripts": [
                (
                    USER_PROVIDED_SCRIPTS_SERVE_URL_PREFIX
                    / umd_scripts_path.name
                    / module_file.name
                ).as_posix()
                for module_file in umd_scripts_path.glob("*.js")
            ],
        }
    )


def setup(server, **kargs):
    client_type = "vue2"
    if hasattr(server, "client_type"):
        client_type = server.client_type

    if client_type == "vue2":
        from . import vue2

        server.enable_module(vue2)
    elif client_type == "vue3":
        from . import vue3

        server.enable_module(vue3)
        setup_handler_module(server)
    else:
        raise TypeError(
            f"Trying to initialize trame_client with unknown client_type={client_type}"
        )



import pytest
from pathlib import Path
from trame_client.utils.testing import FixtureHelper

ROOT_PATH = Path(__file__).parent.parent.absolute()
HELPER = FixtureHelper(ROOT_PATH)


@pytest.fixture
def ref_dir() -> Path:
    return Path(__file__).parent / "refs"


@pytest.fixture
def server(xprocess, server_path):
    name, Starter, Monitor = HELPER.get_xprocess_args(server_path)

    # ensure process is running and return its logfile
    logfile = xprocess.ensure(name, Starter)
    yield Monitor(logfile[1])

    # clean up whole process tree afterwards
    xprocess.getinfo(name).terminate()



from trame.app import get_server
from trame.widgets import html, client
from trame.ui.html import DivLayout

server = get_server(client_type="vue3")
state, ctrl = server.state, server.controller

dyna_var_count = 1


def revert_message():
    state.message = state.message[::-1]


def add_new_var():
    global dyna_var_count
    state[f"new_var_{dyna_var_count}"] = "hello"
    dyna_var_count += 1


# Default UI
with DivLayout(server):
    client.Style(".myRedBG { background: red; } body { margin: 0; }")
    client.ServerTemplate(
        name="loading",
        style="width: 50vw; height: 100vh; position: absolute; left: 0%;",
    )
    client.ServerTemplate(
        name="redSide",
        style="width: 50vw; height: 100vh; position: absolute; left: 50%;",
    )


# ?ui=loading
with DivLayout(server, template_name="loading"):
    client.Loading(message="Yes the loading guy...")

# ?ui=redSide
with DivLayout(server, template_name="redSide"):
    with html.Div(
        "The red side...", classes="myRedBG", style="width: 100%; height: 100%;"
    ):
        eval_js = client.JSEval(
            event=("message", "hello world"),
            exec="window.alert($event)",
        )
        ctrl.exec = eval_js.exec
        html.Div("{{ message }}")
        html.Button("Exec", click=ctrl.exec)
        html.Button("Exec with arg", click=(ctrl.exec, "['Yes me']"))
        html.Button("Change message", click=revert_message)
        html.Br()
        html.Button("New var", click=add_new_var)

server.start()



import time
from trame.app import get_server
from trame.widgets import vuetify3, client, html
from trame.ui.vuetify3 import SinglePageLayout
from trame.ui.html import DivLayout

server = get_server()
server.client_type = "vue3"
state, ctrl = server.state, server.controller

state.trame__title = "Menu example"


def add_element():
    state.items += [f"New elem {time.time()}"]
    state.dirty("items")
    print("add_element", state.items)


@ctrl.add("on_server_reload")
def reload_internal(*args, **kwargs):
    with DivLayout(server, template_name="internal"):
        vuetify3.VSelect(
            v_model=("active_item", "one"), items=("items", ["one", "two", "three"])
        )
        html.Pre("{{ items }}")


reload_internal()

with SinglePageLayout(server) as layout:
    with layout.toolbar:
        vuetify3.VSpacer()
        vuetify3.VBtn("Add", click=add_element)
        vuetify3.VSelect(
            v_model=("active_item", "one"), items=("items", ["one", "two", "three"])
        )
    with layout.content:
        client.ServerTemplate(name="internal")


if __name__ == "__main__":
    server.start()



from trame.app import get_server

server = get_server(client_type="vue2")
ctrl = server.controller
server.state.trame__template_main = "Life Cycle App"


@ctrl.add("on_server_ready")
def server_ready(**state):
    print("on_server_ready")


@ctrl.add("on_client_connected")
def client_connected():
    print("on_client_connected")


@ctrl.add("on_client_unmounted")
def client_unmounted():
    print("on_client_unmounted")


@ctrl.add("on_client_exited")
def client_exited():
    print("on_client_exited")


@ctrl.add("on_server_exited")
def server_exited(**state):
    print("on_server_exited")


server.start()



from pathlib import Path
import logging
import shutil
from dataclasses import dataclass
from urllib.parse import urlparse
from .utils.web_module import file_digest

from .module import (
    USER_PROVIDED_SCRIPTS_DIR_PATH,
    USER_PROVIDED_SCRIPTS_SERVE_URL_PREFIX,
    USER_PROVIDED_UMD_SCRIPTS_DIR_PATH,
    USER_PROVIDED_ES_SCRIPTS_DIR_PATH,
)

logger = logging.getLogger(__name__)


def clean_user_provided_scripts(folder: Path):
    files_to_keep = {"__init__.py", ".gitignore"}

    for item in folder.iterdir():
        if item.is_file() and item.name not in files_to_keep:
            item.unlink()

        if item.is_dir() and item.name not in files_to_keep:
            clean_user_provided_scripts(item)


@dataclass
class UserDefinedFunction:
    name: str
    script: "ExternalScript"


@dataclass
class ExternalScript:
    name: str
    script_original_file_path: Path | str
    function_names: list[str]
    is_umd_module: bool
    umd_global_var_name: str | None
    namespaced_script_file_path: Path | None = None

    def __post_init__(self):
        if isinstance(self.script_original_file_path, str):
            parsed = urlparse(self.script_original_file_path)

            if not all([parsed.scheme, parsed.netloc]):
                msg = f"String must be a valid URL. Received: '{self.script_original_file_path}'"
                raise ValueError(msg)
        else:
            dest_base_path = (
                USER_PROVIDED_UMD_SCRIPTS_DIR_PATH
                if self.is_umd_module
                else USER_PROVIDED_ES_SCRIPTS_DIR_PATH
            )

            script_file_hash = file_digest(self.script_original_file_path)

            self.namespaced_script_file_path = Path(
                f"{script_file_hash}_{self.script_original_file_path.name}"
            )

            shutil.copy2(
                self.script_original_file_path,
                dest_base_path / self.namespaced_script_file_path,
            )

    @property
    def module_path(self) -> str:
        if self.is_umd_module:
            return self.umd_global_var_name

        if isinstance(self.script_original_file_path, str):
            # We have a URL, let's return it as-is
            return self.script_original_file_path

        url_prefix = USER_PROVIDED_SCRIPTS_SERVE_URL_PREFIX
        es_module_prefix = USER_PROVIDED_ES_SCRIPTS_DIR_PATH.name

        return (
            url_prefix / es_module_prefix / self.namespaced_script_file_path
        ).as_posix()

    @property
    def module_type(self) -> str:
        return "umd" if self.is_umd_module else "es"

    def function(self, name: str | None = None) -> UserDefinedFunction:
        if not name:
            try:
                assert len(self.function_names) == 1
            except AssertionError:
                msg = """
                    Ambiguate call to function() without an explicit name.
                    Explicitly specify the function name using the name parameter.
                """
                raise ValueError(msg)

            name = self.function_names[0]

        return UserDefinedFunction(name=name, script=self)


EXTERNAL_SCRIPTS_MAP = {}


def get_external_script_from_name(name: str) -> ExternalScript:
    try:
        return EXTERNAL_SCRIPTS_MAP[name]
    except KeyError as e:
        msg = f"No external script associated with name {name}"
        raise ValueError(msg) from e


def register_user_provided_script(
    script_file_path: Path,
    *,
    function_names: list[str],
    name: str,
    umd: bool = False,
    umd_global_var_name: str | None = None,
) -> ExternalScript:
    if name in EXTERNAL_SCRIPTS_MAP:
        logger.warning("overwriting already registered script %s", name)

    if not EXTERNAL_SCRIPTS_MAP:
        clean_user_provided_scripts(USER_PROVIDED_SCRIPTS_DIR_PATH)

    EXTERNAL_SCRIPTS_MAP[name] = ExternalScript(
        name,
        script_file_path,
        function_names,
        is_umd_module=umd,
        umd_global_var_name=umd_global_var_name,
    )
    return EXTERNAL_SCRIPTS_MAP[name]



from trame.app import get_server
from trame.widgets import html


class MyDiv(html.Div):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._rw = "initial value"

    @property
    def secret(self):
        return "my secret"

    @property
    def rw(self):
        return self._rw

    @rw.setter
    def rw(self, v):
        print("exec setter")
        self._rw = v


def test_component_setter():
    server = get_server("test_component_props")
    w = MyDiv(trame_server=server)
    assert w.secret == "my secret"
    assert w.rw == "initial value"
    w.rw = "hello new value"
    assert w.rw == "hello new value"



from typing import Literal

from trame.app import TrameApp
from trame.ui.vuetify3 import SinglePageLayout

from trame.widgets import client, vuetify3


class ItkWasmPreprocessor(TrameApp):
    def __init__(self, server=None):
        super().__init__(server)

        self.read_dicom_tags_preprocessor_logic = client.register_external_script(
            name="read_dicom_tags",
            script_file_path="https://cdn.jsdelivr.net/npm/@itk-wasm/dicom@7.6.4/dist/bundle/index-worker-embedded.min.js",
            function_names=["readDicomTags"],
        )

        self._build_ui()

    def _build_ui(self):
        with SinglePageLayout(self.server) as ui:
            ui.title.set_text("Client Preprocessing Minimal Example")

            with (
                ui.content,
                vuetify3.VContainer(fluid=True),
                vuetify3.VRow(),
                vuetify3.VCol(cols=3),
                client.Handler(
                    variable="dicom_file",
                    function=self.read_dicom_tags_preprocessor_logic,
                    inputs=("{ tagsToRead: { tags: ['0008|103e'] } }",),
                    completed=(
                        self.preprocessing_pipeline_completed,
                        "[$event.type, $event.outputs.tags]",
                    ),
                ),
            ):
                vuetify3.VFileInput(
                    v_model="dicom_file.value",
                    label="DICOM file",
                    hint="Select a DICOM file to extract tags from",
                    persistent_hint=True,
                )

    def preprocessing_pipeline_completed(
        self, type: Literal["success", "failure"], output
    ) -> None:
        print("preprocessing_pipeline_completed")
        print(type)
        print(output)


def main(**kwargs):
    app = ItkWasmPreprocessor()
    app.server.start(**kwargs)


if __name__ == "__main__":
    main()



class BgColors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


COLOR_PALETTE = [
    BgColors.OKBLUE,
    BgColors.WARNING,
    BgColors.OKCYAN,
    BgColors.FAIL,
    BgColors.OKGREEN,
]


def compute_indent(line: str, increment=2) -> str:
    if line.startswith("</"):
        return -increment
    if line.endswith("/>"):
        return 0
    if line.startswith("<"):
        return increment
    return 0


def to_pretty_html(html_content: str) -> str:
    indent_step = 2
    output_lines = []
    indent = 0
    for line in html_content.splitlines():
        delta = compute_indent(line, indent_step)
        if delta < 0:
            indent += compute_indent(line)

        color = COLOR_PALETTE[int(indent / indent_step) % len(COLOR_PALETTE)]

        output_lines.append(
            f"{color}{' ' * indent}{line.replace(' >', '>')}{BgColors.ENDC}"
        )
        if delta > 0:
            indent += compute_indent(line)

    return "\n".join(output_lines)



from trame.app import get_server
from trame.ui.html import DivLayout
from trame.widgets import html

server = get_server(client_type="vue3")

# Test to dynamically add a fake directive
html.Div.register_directive("v_seb_directive")

with DivLayout(server) as layout:
    with html.Div(v_seb_directive=True):
        with html.Div():
            with html.Div(
                style="border: 1px;",
                classes=("dynaClass", {}),
            ) as sub_container:
                html.Span("Hello")

            print("=" * 60)
            print("widget")
            print("=" * 60)
            print(sub_container)

            html.Span("world")
        html.Span("What's")
    html.Span("Up")

    print("=" * 60)
    print("Full layout")
    print("=" * 60)
    print(layout)



"""
Trame example that checks an input DICOM file patient tag against a patient name in the trame state.
"""

from pathlib import Path
from typing import Literal

from trame.app import TrameApp
from trame.ui.vuetify3 import SinglePageLayout

from trame.widgets import client, vuetify3

DICOM_UTILS_SCRIPT = Path(__file__).with_name("dicom_utils.js")


class DicomProcessingPipeline(TrameApp):
    def __init__(self, server=None):
        super().__init__(server)

        self.state.patient_name = "John Doe"

        # Mark your state as client_only!
        self.state.client_only("dicom_file_set")

        self.dicom_utils_preprocessor = client.register_external_script(
            DICOM_UTILS_SCRIPT,
            name="dicom_utils",
            function_names=[
                "validateDicomFilePatientName",
                "generateNIFTIFromDicomFileSet",
            ],
        )

        self._build_ui()

    def _build_ui(self):
        with SinglePageLayout(self.server) as ui:
            ui.title.set_text("Nested Pipeline")

            with (
                ui.content,
                vuetify3.VContainer(fluid=True),
                vuetify3.VRow(),
                vuetify3.VCol(cols=3),
                client.Handler(
                    variable="validated_dicom_file",
                    function=("dicom_utils", "generateNIFTIFromDicomFileSet"),
                    inputs=("{ output_file_name: 'my_converted_file.nii' }",),
                    completed=(
                        self.preprocessing_pipeline_completed,
                        "[$event.type, $event.outputs]",
                    ),
                ),
                client.Handler(
                    variable="input_dicom_fileset",
                    function=self.dicom_utils_preprocessor.function(
                        "validateDicomFilePatientName"
                    ),
                    inputs=("{ patient_name }",),
                    success="validated_dicom_file.value = $event",
                    failure=(self.on_patient_name_failure, "[$event]"),
                ),
            ):
                vuetify3.VFileInput(
                    v_model="input_dicom_fileset.value",
                    label="DICOM files",
                    hint="""
                        Input here a single DICOM file that will be patient name checked
                        and then converted to a single NIFTI file.
                    """,
                    persistent_hint=True,
                )
                vuetify3.VTextField(
                    v_model="patient_name",
                    hint="Input here the patient name that you expect for your DICOM file",
                    label="Expected Patient Name",
                    persistent_hint=True,
                )

                vuetify3.VAlert(
                    v_if="alert_title",
                    text=("alert_text", None),
                    title=("alert_title", None),
                    type=("alert_type", None),
                )

    def on_patient_name_failure(self, output):
        print(output)
        self.state.alert_type = "error"
        self.state.alert_title = self.state.alert_type
        self.state.alert_text = (
            f"Expected patient {output[0]['expected']}, got {output[0]['got']}"
        )

    def preprocessing_pipeline_completed(
        self, type: Literal["success", "failure"], output
    ) -> None:
        print("preprocessing_pipeline_completed")
        print(type)
        print(output)

        self.state.alert_type = type if type == "success" else "error"
        self.state.alert_title = self.state.alert_type

        if type == "failure":
            self.state.alert_text = "Failed to convert to NIFTI file."
        else:
            self.state.alert_text = (
                "Patient name matched! We got the converted NIFTI file."
            )


def main(**kwargs):
    app = DicomProcessingPipeline()
    app.server.start(**kwargs)


if __name__ == "__main__":
    main()



import pytest

from playwright.sync_api import expect
from trame_client.utils.testing import assert_snapshot_matches


@pytest.mark.parametrize(
    "server_path",
    ["examples/vue2/dynamic_template.py"],
)
def test_dynamic_template(server, page, ref_dir):
    url = f"http://127.0.0.1:{server.port}/"
    page.goto(url)

    assert_snapshot_matches(page, ref_dir, "test_dynamic_template_initial")

    expect(page.locator(".staticDiv")).to_have_text("Static text 2")
    expect(page.locator(".countDiv")).to_have_text("count = 2")
    expect(page.locator(".ttsDiv")).to_have_text("tts = 0")
    assert server.get("count") == 2

    page.locator(".plusBtn").click()
    page.locator(".plusBtn").click()
    expect(page.locator(".staticDiv")).to_have_text("Static text 2")
    assert server.get("count") == 4
    expect(page.locator(".countDiv")).to_have_text("count = 4")
    expect(page.locator(".ttsDiv")).to_have_text("tts = 0")
    page.locator(".updateBtn").click()
    page.locator(".updateBtn").click()
    expect(page.locator(".staticDiv")).to_have_text("Static text 6")
    assert server.get("count") == 6
    expect(page.locator(".countDiv")).to_have_text("count = 6")
    expect(page.locator(".ttsDiv")).to_have_text("tts = 2")

    assert_snapshot_matches(page, ref_dir, "test_dynamic_template_final")


@pytest.mark.parametrize(
    "server_path",
    ["examples/vue2/js_call.py", "examples/vue3/js_call.py"],
)
def test_js_call(server, page):
    url = f"http://127.0.0.1:{server.port}/"
    page.goto(url)

    expect(page.locator(".jsAlert")).to_have_text("Alert")
    assert server.get("message") == "hello world"
    page.locator(".alertMsg").click()
    expect(page.locator(".jsAlert")).to_have_text("hello world")
    page.locator(".alertMe").click()
    expect(page.locator(".jsAlert")).to_have_text("Yes me")
    page.locator(".swapMsg").click()
    assert server.get("message") == "dlrow olleh"
    page.locator(".alertMsg").click()
    expect(page.locator(".jsAlert")).to_have_text("dlrow olleh")



from trame.app import get_server

server = get_server(client_type="vue2")
state, ctrl = server.state, server.controller

state.count = 2
state.double = 4
state.client_type = server.client_type


@state.change("count")
def update_count(count, **kwargs):
    state.double = 2 * int(count)


state.trame__template_main = """
    <div>
        <label>{{ client_type }}</label>
        <div>count = {{ count }}</div>
        <div>2 x count = {{ utils.fmt.bytes(double) }}</div>
        <input type="range" min="0" max="1000000" step="1000" v-model="count" />
    </div>
"""

server.start()



from trame.app import get_server
from trame.widgets import html, client
from trame.ui.html import DivLayout

server = get_server()

JS_CONTENT = (
    "import confetti from 'https://esm.sh/canvas-confetti';"
    "trame.utils.confetti = confetti;"
)

with DivLayout(server):
    html.Button("Click Me", click="utils.confetti()")
    client.Script(JS_CONTENT, module=True)

server.start()



from trame.app import get_server
from trame.widgets import client, html
from trame.ui.html import DivLayout

server = get_server(client_type="vue2")

state = server.state
global_layout = None

state.a = 1
state.last_event = None


def add():
    state.a += 1


def add_div():
    with global_layout:
        html.Div("Hello")


with DivLayout(server) as layout:
    global_layout = layout
    layout.root.add_child(
        "a={{ a }} - LifeCycle {{ last_event }}<br> template_ts={{ tts }} <br> "
    )
    html.Button("a+", click="a+=1")
    html.Button("set(a+)", click="set('a', a+1)")
    html.Button("server", click=add)
    html.Button("modify layout", click=add_div)
    client.LifeCycleMonitor(
        type="emit",
        created="last_event += 'created '",
        before_mount="last_event += 'beforeMount '",
        mounted="last_event += 'mounted '",
        before_update="last_event += 'beforeUpdate' ",
        updated="last_event += 'updated '",
        before_destroy="last_event += 'beforeDestroy '",
        destroyed="last_event += 'destroyed '",
    )


if __name__ == "__main__":
    server.start()



import hashlib
from pathlib import Path


def file_digest(file_path):
    """Return the hex digest of a file"""
    h = hashlib.sha256()
    b = bytearray(128 * 1024)
    mv = memoryview(b)
    with open(file_path, "rb", buffering=0) as f:
        while n := f.readinto(mv):
            h.update(mv[:n])
    return h.hexdigest()


def file_with_digest(file_path, digest=None, digest_size=40):
    """Write a new file sibbling to file_path with a digest hash in his name"""
    file_path = Path(file_path)

    if digest is None:
        digest = file_digest(file_path)

    if len(digest) > digest_size:
        digest = digest[:digest_size]

    # Create file with digest name
    output_file = file_path.with_name(f"{file_path.stem}-{digest}{file_path.suffix}")
    output_file.write_bytes(file_path.read_bytes())

    return output_file


def is_relative_to(path, *other_paths):
    """Helper to check that *other_paths are nested under path"""
    # Convert the input path and the base path into absolute paths
    abs_path = path.resolve()
    abs_other_paths = [Path(other).resolve() for other in other_paths]

    # Check if the parts of the base paths are the beginning of the input path's parts
    return all(
        abs_path.parts[: len(base.parts)] == base.parts for base in abs_other_paths
    )


class WebModule:
    """Helper class to create/define a module while dynamically computing hash
    to properly prevent invalid browser cache.
    """

    def __init__(self, vue_use=None, digest_size=6):
        self._digest_size = digest_size
        self._serving_entries = []
        self._serve = {}
        self._vue_use = []
        self._styles = []
        self._scripts = []

        if vue_use is not None:
            self._vue_use.append(vue_use)

    def _add_file(self, file_path):
        file_path = Path(file_path).resolve()
        if not file_path.exists():
            raise ValueError(f"File {file_path} does not exist")

        for entry in self._serving_entries:
            fs_path, www_path = entry
            if is_relative_to(file_path, fs_path):
                if self._digest_size > 0:
                    file_path = file_with_digest(
                        file_path, digest_size=self._digest_size
                    )

                r_path = file_path.relative_to(fs_path)
                return f"{www_path}/{r_path}"

        raise ValueError(f"No parent found for {file_path}")

    def serve_directory(self, directory_to_serve, www_base_name):
        """Register a new directory to serve with its alias endpoint"""
        fs_path = Path(directory_to_serve).resolve()
        if not fs_path.exists():
            raise ValueError(
                f"Directory {fs_path} does not exist - Won't be able to serve it"
            )
        if www_base_name in self._serve:
            raise ValueError(
                f"Entry for {www_base_name} is already set for {self._serve[www_base_name]}"
                f"and want to override with {fs_path}"
            )

        self._serving_entries.append((fs_path, www_base_name))
        self._serve[www_base_name] = str(fs_path)

    def add_script_file(self, file_path):
        """Add a script file to serve while computing unique name and proper http path internally."""
        self._scripts.append(self._add_file(file_path))

    def add_style_file(self, file_path):
        """Add a css file to serve while computing unique name and proper http path internally."""
        self._styles.append(self._add_file(file_path))

    def add_vue_use(self, vue_use):
        """Register a vue plugin expression"""
        self._vue_use.append(vue_use)

    @property
    def module(self):
        """Return the generated trame module as a dictionary."""
        return dict(
            serve=self._serve,
            scripts=self._scripts,
            styles=self._styles,
            vue_use=self._vue_use,
        )



import os
from trame.app import get_server
from trame.widgets import html
from trame.ui.html import DivLayout
import time

server = get_server(client_type=os.environ.get("TRAME_CLIENT_TYPE", "vue3"))

# -----------------------------------------------------------------------------
# Only with Vue3
# -----------------------------------------------------------------------------
with DivLayout(server, template_name="error") as layout:
    html.Div("App disconnected: {{ trame_error_report_msg }}")

with DivLayout(server, template_name="error_reconnect") as layout:
    html.Div("App disconnected but you can reconnect: {{ trame_error_report_msg }}")
# -----------------------------------------------------------------------------

with DivLayout(server) as layout:
    html.Div("App is fully loaded")

# Busy wait
time.sleep(5)

server.start()



from playwright.sync_api import expect
import pytest

from trame_client.utils.testing import assert_snapshot_matches


@pytest.mark.parametrize("server_path", ["examples/test/reactivity.py"])
def test_reactivity(server, page, ref_dir):
    url = f"http://127.0.0.1:{server.port}/"
    page.goto(url)

    assert_snapshot_matches(page, ref_dir, "simple_count_1")

    plus_button = page.locator(".plusButton")
    count_value = page.locator(".countValue")

    expect(count_value).to_have_text("1")
    assert server.get("count") == 1
    plus_button.click()
    expect(count_value).to_have_text("2")
    assert server.get("count") == 2
    plus_button.click()
    plus_button.click()
    expect(count_value).to_have_text("4")
    assert server.get("count") == 4
    plus_button.click()
    expect(count_value).to_have_text("5")
    assert server.get("count") == 5

    assert_snapshot_matches(page, ref_dir, "simple_count_5")



from trame.app import get_server
from trame.widgets import html
from trame.ui.html import DivLayout

server = get_server()

with DivLayout(server) as layout:
    html.Div("App is running")

server.start()



# Run with -d or --debug
from trame.app import get_server
from trame.ui.html import DivLayout
from trame.widgets import html

server = get_server(client_type="vue3")

with DivLayout(server) as layout:
    html.Div(
        v_show=True,
        test=({},),
        test_1=([],),
        test_2=(5,),
        __properties=[
            "test",
            "test_1",
            "test_2",
        ],
    )



from trame.app import get_server
from trame.ui.html import DivLayout
from trame.widgets import client, html

server = get_server(client_type="vue3")


@server.state.change("first_size")
def on_size(first_size, **kwargs):
    print(first_size)


with DivLayout(server) as layout:
    layout.root.style = "height: 100vh; width: 100%; padding: 0; margin: 0;"

    with html.Div(
        style="width: 40vw; height: 40vh; border: solid 1px #333; margin: 25px;"
    ):
        with client.SizeObserver("first_size"):
            html.Div(
                "{{ first_size?.size.width.toFixed(1) }} x {{ first_size?.size.height.toFixed(1) }}",
                style="margin: 0 auto; height: 100%; text-align: middle;",
            )

    with html.Div(
        style="width: 20vw; height: 30vh; border: solid 1px #333; position: absolute; right: 10px; bottom: 10px"
    ):
        with client.SizeObserver("second_size"):
            html.Div(
                "{{ second_size?.size.width.toFixed(1) }} x {{ second_size?.size.height.toFixed(1) }}",
                style="margin: 0 auto; height: 100%; text-align: middle;",
            )


if __name__ == "__main__":
    server.start()



import logging
import uuid
from pathlib import Path

from .core import AbstractElement
from ..external_script_handler import (
    UserDefinedFunction,
    ExternalScript,
    get_external_script_from_name,
    register_user_provided_script,
)

logger = logging.getLogger(__name__)

__all__ = [
    "Loading",
    "ServerTemplate",
    "Script",
    "Style",
    "JSEval",
    "Getter",
    "ClientStateChange",
    "ClientTriggers",
    "DeepReactive",
    "LifeCycleMonitor",
    "SizeObserver",
    "Handler",
    "register_external_script",
]


# -----------------------------------------------------------------------------
# TrameLoading
# -----------------------------------------------------------------------------
class Loading(AbstractElement):
    """
    Three circle spinning with a message

    :param message: Content of the message
    :type name: str
    """

    def __init__(self, children=None, **kwargs):
        super().__init__("trame-loading", children, **kwargs)
        self._attr_names += [
            "message",
        ]


# -----------------------------------------------------------------------------
# TrameServerTemplate
# -----------------------------------------------------------------------------
class ServerTemplate(AbstractElement):
    """
    Template content presentation

    :param name: Name of the template to display. (default: 'main')
    :type name: str

    :param use_url: If true, will use the `{url_key}=` from the url to override
                    the `name=` attribute.
    :type use_url: bool

    :param url_key: Name of the key to extract from url for use_url. (default: 'ui')
    :type url_key: str
    """

    def __init__(self, children=None, **kwargs):
        super().__init__("trame-template", children, **kwargs)
        self._attr_names += [
            ("name", "templateName"),
            ("use_url", "useUrl"),
            ("url_key", "urlKey"),
        ]


# -----------------------------------------------------------------------------
# TrameExec
# -----------------------------------------------------------------------------
class JSEval(AbstractElement):
    """Provide means to execute JS code"""

    _next_id = 0

    def __init__(self, children=None, **kwargs):
        super().__init__("trame-exec", children, **kwargs)
        JSEval._next_id += 1
        self.__ref = kwargs.get("ref", f"trame_exec_ref_{JSEval._next_id}")
        self._attributes["ref"] = f'ref="{self.__ref}"'
        self._attr_names += [
            "event",
        ]
        self._event_names += [
            "exec",
        ]

    def exec(self, event=None):
        if event is None:
            self.server.js_call(self.__ref, "exec")
        else:
            self.server.js_call(self.__ref, "exec", event)


# -----------------------------------------------------------------------------
# TrameStyle
# -----------------------------------------------------------------------------
class Style(AbstractElement):
    """Provide means to inject global CSS rules"""

    _next_id = 0

    def __init__(self, css_content=None, **kwargs):
        Style._next_id += 1
        super().__init__("trame-style", **kwargs)
        self._key = f"trame__inline_style_{Style._next_id}"
        self._attributes["_css"] = f':css="{self._key}"'
        self.server.root_server.state.setdefault(self._key, css_content)

    def update(self, css_content):
        """Update style content"""
        self.server.root_server.state[self._key] = css_content

    @property
    def var_name(self):
        """Name the the state variable used by this widget"""
        return self._key


# -----------------------------------------------------------------------------
# TrameScript
# -----------------------------------------------------------------------------
class Script(AbstractElement):
    """Provide means to inject a global script tag"""

    _next_id = 0

    def __init__(self, script_content=None, **kwargs):
        Script._next_id += 1
        super().__init__("trame-script", **kwargs)
        self._key = f"trame__inline_script_{Script._next_id}"
        self._attributes["_script"] = f':script="{self._key}"'
        self.server.root_server.state.setdefault(self._key, script_content)
        self._attr_names += [
            "module",  # type="module" or "text/javascript"
            "async",
            "crossorigin",
            "defer",
            "integrity",
            "nomodule",
            "referrerpolicy",
            "src",
        ]

    def update(self, script_content):
        """Update style content"""
        self.server.root_server.state[self._key] = script_content

    @property
    def var_name(self):
        """Name the the state variable used by this widget"""
        return self._key


# -----------------------------------------------------------------------------
# TrameGetter
# -----------------------------------------------------------------------------
class Getter(AbstractElement):
    """Provide means to extract a reactive state variable from its name

    :param name: Name of the state variable to extract. This can be an expression too.

    :param value_name: Name of the JavaScript variable name (default: "value")
    :type value_name: str

    :param key_name: Name for the JavaScript variable that will hold the evaluate expression of the "name" property.

    :param update_nested_name: Method name if you aim to update the nested structure.

    :param update_name: Method name if you aim to update the full value.
    """

    def __init__(
        self,
        children=None,
        value_name=None,
        key_name=None,
        update_nested_name=None,
        update_name=None,
        **kwargs,
    ):
        super().__init__("trame-getter", **kwargs)
        self._attr_names += [
            "name",
        ]
        extracts = []
        if key_name:
            extracts.append(f"keyName: {key_name}")

        if update_nested_name:
            extracts.append(f"updateNested: {update_nested_name}")

        if update_name:
            extracts.append(f"update: {update_name}")

        if value_name:
            extracts.append(f"value: {value_name}")
        else:
            extracts.append("value")

        self._attributes["slot"] = f'v-slot="{{ {", ".join(extracts)} }}"'


# -----------------------------------------------------------------------------
# TrameClientStateChange
# -----------------------------------------------------------------------------
class ClientStateChange(AbstractElement):
    """
    Allow the client side to trigger an event when a state element changes.

    :param value: Name of the state variable to monitor
    :type value: str
    :param immediate: Trigger change right away rather than at nextTick (default: False)
    :type immediate: bool
    :param trigger_on_create: If set to true, the change event will be triggered when the client start. (default: False)
    :type trigger_on_create: bool
    :param change: Event triggered when state[value] change
    :type change: Function or JS expression (event)
    """

    def __init__(self, children=None, **kwargs):
        super().__init__("trame-client-state-change", children, **kwargs)
        self._attr_names += [
            ("value", ":value"),
            "immediate",
            ("trigger_on_create", "triggerChangeOnCreate"),
        ]
        self._event_names += ["change"]


# -----------------------------------------------------------------------------
# TrameClientTriggers
# -----------------------------------------------------------------------------
class ClientTriggers(AbstractElement):
    """
    Allow to trigger an event on the client side

    :param ref: Identifier for the client side DOM elem
    :param **kwargs: List of events to registers with their callbacks

    Built-in events are:
       - mounted
       - created
       - before_destroy
       - before_unmount
       - exit
    """

    def __init__(self, ref="trame_triggers", children=None, **kwargs):
        self.__name = ref
        super().__init__("trame-client-triggers", children=None, ref=ref, **kwargs)
        self._attr_names += ["ref"]
        self._event_names += [
            ("before_destroy", "beforeDestroy"),
            ("before_unmount", "beforeUnmount"),
        ]
        self._event_names += list(kwargs.keys())

    def call(self, method, *args):
        """
        Perform the call on the client

        :param method: Key used in the kwargs at construction time
        """
        self.server.js_call(self.__name, "emit", method, *args)


# -----------------------------------------------------------------------------
# TrameDeepReactive
# -----------------------------------------------------------------------------
class DeepReactive(AbstractElement):
    """Create a deeply reactive state from state variable name.
    The provided name can not be reactive.
    It needs to be statically defined in Python like in the example blow.

    This component only works with client_type="vue3".

       with DeepReactive(my_nested_var):
          html.Input(v_model=my_nested_var.txt_1)
          html.Input(v_model=my_nested_var.txt_2)
    """

    def __init__(
        self,
        name=None,
        children=None,
        **kwargs,
    ):
        super().__init__("trame-deep-reactive", children, name=name, **kwargs)
        self._attr_names += [
            "name",
        ]

        self._attributes["slot"] = f'v-slot="{{ value: {name} }}"'


# -----------------------------------------------------------------------------
# TrameLifeCycleMonitor
# -----------------------------------------------------------------------------
class LifeCycleMonitor(AbstractElement):
    """
    LifeCycleMonitor is a debug purpose tool to validate a sub-tree get the proper
    expected life cycle event.

    This component allow to print into the client side console when any of the
    monitored event happen.

    :param name: User specific text to easily identify which component the event
                was coming from.
    :param type: console[type](...) so you can use 'log', 'error', 'info', 'warn'
    :param value: Another value that is printed when an event occur
    :param events: List of events to monitor such as created, beforeMount,
        mounted, beforeUpdate, updated, beforeDestroy, destroyed
    """

    def __init__(self, children=None, **kwargs):
        super().__init__("trame-life-cycle-monitor", children, **kwargs)
        self._attr_names += [
            "name",
            "type",
            "value",
            "events",
        ]
        self._event_names += [
            "created",
            ("before_mount", "beforeMount"),
            "mounted",
            ("before_update", "beforeUpdate"),
            "updated",
            ("before_destroy", "beforeDestroy"),
            "destroyed",
        ]


# -----------------------------------------------------------------------------
# TrameSizeObserver
# -----------------------------------------------------------------------------
class SizeObserver(AbstractElement):
    """
    SizeObserver allow to monitor the space available in the UI and bind that
    information onto a state variable.

    :param _name: Name of the state variable to bound the component size to
    """

    def __init__(self, _name, **kwargs):
        super().__init__("trame-size-observer", **kwargs)
        self._attr_names += [
            "name",
        ]
        self.server.state[_name] = None
        self.name = self.server.translator.translate_key(_name)


def register_external_script(
    script_file_path: Path | str,
    *,
    function_names: list[str],
    name: str | None = None,
    umd: bool = False,
    umd_global_var_name: str | None = None,
) -> ExternalScript:
    """
    Register an external script that can be used together with the Handler widget.

    Args:
        script_file_path (Path | str):
            The local of the script you want to register. Can be either a local or a remote script. Accepts:
            - `Path`: local path to your script on the filesystem
            - `str`: an valid URL to the script
        function_names (list[str]): public API of your script that you want to expose to the Handler widget.
            Accepts a list of function names that must be defined in your script.
        name (str): a registration identifier for your script.
            You may use it later to reference your script when instantiating the Handler widget.
        umd (bool, optional): If true, your script is considered as a UMD module and treated as such.
            For ES module, just keep the default value. Defaults to False.
        umd_global_var_name (str | None):
            Specify the global variable name the UMD module will use to register itself on the browser window object.
            Required if and only if `umd` is set to True. Defaults to None.
    """

    # TODO - support inline scripts without functions to be called
    assert len(function_names) > 0

    if not name:
        name = str(uuid.uuid4())

    return register_user_provided_script(
        script_file_path,
        function_names=function_names,
        name=name,
        umd=umd,
        umd_global_var_name=umd_global_var_name,
    )


# -----------------------------------------------------------------------------
# Handler
# -----------------------------------------------------------------------------
class Handler(AbstractElement):
    """
    Facilitates the execution of registered JS functions and external scripts.

    The Handler simplifies calling client-side logic registered via `register_external_script`.
    It manages the execution flow between the Python state and JavaScript functions,
    providing events to monitor execution status.

    Args:
        function (UserDefinedFunction | tuple[str, str] | ExternalScript | str):
            The JavaScript function to execute. Accepts:
            - `UserDefinedFunction`: Obtained from `ExternalScript.function`.
            - `tuple[str, str]`: A `(module_name, function_name)` pair.
            - `ExternalScript`: An object from `register_external_script` (must
              contain exactly one function).
            - `str`: The registration identifier of an external script.
        variable (str): The name of the state variable to be passed as the
            primary input to the function.
        trigger_on_change (bool, optional): If True, the function is automatically
            invoked whenever `variable` changes. Defaults to True.
        inputs (dict, optional): A dictionary of additional key-value pairs
            to be passed as arguments to the function. Defaults to None.

    Events:
        completed: A generic event emitted regardless of the execution outcome.
        success: Emitted when the function executes and returns successfully.
        failure: Emitted when the function reports a handled error.
        error: Emitted when an unhandled exception occurs during execution.
    """

    def __init__(
        self,
        *,
        function: UserDefinedFunction | tuple[str, str] | ExternalScript | str,
        variable: str = "input_data",
        **kwargs,
    ):
        external_script_id = f"trame_external_script{id(self)}"

        super().__init__("trame-handler", **kwargs, function_key=external_script_id)

        self.variable = variable

        if not self.state["trame__client_external_scripts"]:
            self.state["trame__client_external_scripts"] = []

        if isinstance(function, UserDefinedFunction):
            external_script = function.script
            function_handle = function

        if isinstance(function, tuple):
            external_script = get_external_script_from_name(function[0])
            function_handle = external_script.function(function[1])

        if isinstance(function, ExternalScript):
            external_script = function
            function_handle = external_script.function()

        if isinstance(function, str):
            external_script = get_external_script_from_name(function)
            function_handle = external_script.function()

        self.state["trame__client_external_scripts"].append(
            {
                "module_type": external_script.module_type,
                "path": external_script.module_path,
                "function": function_handle.name,
                "id": external_script_id,
            }
        )

        self._attr_names += [
            "trigger_on_change",
            "inputs",
            ("function_key", "functionKey"),
        ]
        self._event_names += ["success", "failure", "error", "completed"]

        self._attributes["slot"] = (
            f'v-slot:default="{{ input_data: {self.variable}, run }}"'
        )

    @property
    def input_data(self) -> str:
        """
        Get the JS expression that references the handler's input data.
        """
        return f"{self.variable}.value"

    def run(self, input_var: str) -> str:
        """
        Return a JS expression that will invoke the handler with the provided input_var as argument.
        """
        return f"run({input_var})"



from trame.app import get_server
from trame_client.widgets.html import Div

CLIENT_TYPE = "vue3"


def test_properties():
    server = get_server("test_properties", client_type=CLIENT_TYPE)
    widget = Div(
        trame_server=server,
        v_model_hello="world",
        v_model_hello_world="something",
        v_bind_hello="world",
        v_bind_hello_world="something",
        v_model_number="number_value",
    )
    assert widget.html == (
        "<div "
        'v-model:hello="world" '
        'v-model:hello.world="something" '
        ':hello="world" '
        ':hello.world="something" '
        'v-model.number="number_value" '
        "/>"
    )


def test_events():
    server = get_server("test_events", client_type=CLIENT_TYPE)
    widget = Div(
        trame_server=server,
        v_on_click="click",
        v_on_mouseenter_left="left",
        v_on_mouseout_left_stop="left_stop",
    )
    assert widget.html == (
        '<div @click="click" @mouseenter.left="left" @mouseout.left.stop="left_stop" />'
    )



"""
Trame example that uses an UMD module to file the holes of a STL file before visualization.
You need to build the JS project inside geometry-preprocessing folder in order for this to work.
You can download an example file from here: https://kitware.github.io/vtk-js/data/stl/spherewithholes.stl
"""

from pathlib import Path
import tempfile
from typing import Literal

from trame.app import TrameApp
from trame.ui.vuetify3 import SinglePageLayout

from trame.app.file_upload import ClientFile
from trame.widgets import client, vuetify3, vtk as vtk_widgets
from vtkmodules import all as vtk

VTK_UTILS_SCRIPT = (
    Path(__file__).with_name("geometry-preprocessing")
    / "dist"
    / "geometry-preprocessing.umd.js"
)


class PatientNameValidator(TrameApp):
    def __init__(self, server=None):
        super().__init__(server)

        self.state.decimation_factor = 5
        self.state.hole_size = 1000.0

        client.register_external_script(
            name="fill_stl_holes",
            script_file_path=VTK_UTILS_SCRIPT,
            function_names=["fill_stl_holes"],
            umd=True,
            umd_global_var_name="GeometryPreProcessing",
        )

        # If your preprocessing depends on some libs, you can register them using server.enable_module
        # self.server.enable_module(
        #     {"scripts": ["https://unpkg.com/@kitware/vtk-wasm/vtk-umd.js"]}
        # )

        self.setup_vtk()
        self._build_ui()

    def setup_vtk(self):
        renderer = vtk.vtkRenderer()
        render_window = vtk.vtkRenderWindow()
        render_window.AddRenderer(renderer)
        render_window.OffScreenRenderingOn()

        render_window_interactor = vtk.vtkRenderWindowInteractor()
        render_window_interactor.SetRenderWindow(render_window)
        render_window_interactor.GetInteractorStyle().SetCurrentStyleToTrackballCamera()

        self.renderer = renderer
        self.render_window = render_window

    def _build_ui(self):
        with SinglePageLayout(self.server) as self.ui:
            self.ui.title.set_text("Client Preprocessing UMD Modules")

            with self.ui.toolbar:
                with client.Handler(
                    function="fill_stl_holes",
                    inputs=("{ hole_size }",),
                    completed=(
                        self.preprocessing_pipeline_completed,
                        "[$event.type, $event.outputs]",
                    ),
                ) as fill_stl_holes:
                    vuetify3.VNumberInput(
                        v_model="hole_size",
                        label="Input here the hole size used to configure the VTK.js filter.",
                    )

                    vuetify3.VFileInput(
                        v_model=fill_stl_holes.input_data,
                        accept=".stl",
                        label="""
                            Input here an STL file with holes that needs to be filled.
                            It will be processed by VTK.js
                        """,
                    )

            with (
                self.ui.content,
                vuetify3.VContainer(fluid=True, classes="pa-0 fill-height"),
            ):
                view = vtk_widgets.VtkRemoteView(self.render_window)
                self.ctrl.view_update = view.update
                self.ctrl.view_reset_camera = view.reset_camera

    def preprocessing_pipeline_completed(
        self, event_type: Literal["success", "failure"], output
    ) -> None:
        if event_type == "failure":
            print("pipeline failed")
            return

        stl_file = ClientFile(output[0])

        with tempfile.NamedTemporaryFile(mode="wb") as f:
            print(type(stl_file.content))
            f.write(stl_file.content)
            f.flush()

            stl_reader = vtk.vtkSTLReader(file_name=f.name)
            mapper = vtk.vtkPolyDataMapper()

            stl_reader >> mapper

            actor = vtk.vtkActor()
            actor.SetMapper(mapper)
            self.renderer.AddActor(actor)
            self.renderer.ResetCamera()
            self.render_window.Render()

            self.ctrl.view_update()


def main(**kwargs):
    app = PatientNameValidator()
    app.server.start(**kwargs)


if __name__ == "__main__":
    main()



from trame.app import get_server
from trame.widgets import client, html
from trame.ui.html import DivLayout

server = get_server(client_type="vue2")


def a_changed():
    print(f"a={server.state.a}")


with DivLayout(server) as layout:
    html.Div("{{ a }}")
    html.Input(type="range", v_model=("a", 1))
    client.ClientStateChange(value="a", change=a_changed)


if __name__ == "__main__":
    server.start()



import random
import numpy as np
from PIL import Image
import pytest

from trame_client.utils.testing import assert_images_match


def test_assert_images_match(ref_dir):
    ref_image_path = ref_dir.joinpath("ref1.png")
    assert ref_image_path.exists()

    image = Image.open(ref_image_path)
    assert_images_match(image, ref_image_path, 0.0)

    W, H = image.size
    target_error = int(0.09 * H * W)
    coordinates = [
        (int(random.random() * H), int(random.random() * W))
        for _ in range(target_error)
    ]
    np_img = np.array(image)
    for y, x in coordinates:
        np_img[y, x] = random.randint(0, 255)
    assert_images_match(Image.fromarray(np_img), ref_image_path, 0.1)

    with pytest.raises(AssertionError):
        assert_images_match(Image.fromarray(np_img), ref_image_path, 0.05)



from pathlib import Path

from trame.app import TrameApp
from trame.app.file_upload import ClientFile
from trame.ui.vuetify3 import SinglePageLayout

from trame.widgets import client, html, vuetify3

DICOM_UTILS_SCRIPT = Path(__file__).with_name("dicom_utils.js")


class DicomToNiftiConvertor(TrameApp):
    def __init__(self, server=None):
        super().__init__(server)

        self.state.dicom_file_set = None
        self.state.feedback_message = ""
        self.state.is_error = False
        self.state.is_success = False

        # Mark your state as client_only!
        self.state.client_only("dicom_file_set")

        client.register_external_script(
            name="generate_nii_from_dcm",
            script_file_path=DICOM_UTILS_SCRIPT,
            function_names=["generateNIFTIFromDicomFileSet"],
        )

        self._build_ui()

    def _build_ui(self):
        with SinglePageLayout(self.server) as self.ui:
            self.ui.title.set_text("Explicit Pipeline Trigger")

            with (
                self.ui.content,
                vuetify3.VContainer(fluid=True),
                vuetify3.VRow(),
                vuetify3.VCol(cols=3),
            ):
                with vuetify3.VCard(flat=True):
                    with (
                        vuetify3.VSnackbar(
                            model_value=("is_error || is_success",),
                            color=("is_error ? 'error' : 'success'",),
                            location="top left",
                            close_on_content_click=True,
                            timeout=6000,
                            update_modelValue="is_error = $event; is_success = $event;",
                        ),
                        html.Div(style="align-items: center;"),
                    ):
                        vuetify3.VIcon("mdi-information", style="margin-right: 12px")
                        html.Span("{{ feedback_message }}")

                    vuetify3.VFileInput(
                        v_model="dicom_file_set",
                        multiple=True,
                        label="DICOM files",
                        hint="Input here a set of DICOM files that will be converted to a single .nii file",
                        persistent_hint=True,
                    )

                    with client.Handler(
                        function="generate_nii_from_dcm",
                        inputs=("{ output_file_name: 'my_converted_file.nii' }",),
                        # success event for the logic happy path
                        success=(self.save_nifti_file, "[$event]"),
                        # failure event for the logic error path (e.g. invalid data)
                        failure=(self.handle_nifti_convertion_error, "[$event]"),
                        # error event for any exception raised during logic execution
                        error=(self.handle_nifti_convertion_error, "[$event]"),
                    ) as client_handler:
                        # trigger the preprocessing on our dicom_file_set state key
                        vuetify3.VBtn(
                            "Convert DICOMs to NIFTI file",
                            color="primary",
                            click=client_handler.run("dicom_file_set"),
                        )

    def save_nifti_file(self, raw_nifti_file):
        nifti_file = ClientFile(raw_nifti_file[0])
        print("got a NIFTI file")

        print(nifti_file.info)
        print(nifti_file.name)

        self.state.feedback_message = "NIFTI file successfully generated"
        self.state.is_success = True
        self.state.is_error = False

        file_path = (Path("outputs") / nifti_file.name).resolve()
        file_path.parent.mkdir(parents=True, exist_ok=True)

        if file_path.exists():
            file_path.unlink()

        with file_path.open(mode="xb") as file:
            file.write(nifti_file.content)

    def handle_nifti_convertion_error(self, error):
        print(error)
        self.state.feedback_message = error
        self.state.is_error = True


def main(**kwargs):
    app = DicomToNiftiConvertor()
    app.server.start(**kwargs)


if __name__ == "__main__":
    main()



from trame.app import get_server
from trame.widgets import html
from trame.ui.html import DivLayout

server = get_server()
server.client_type = "vue2"
state, ctrl = server.state, server.controller

state.count = 2
state.double = 4


@state.change("count")
def update_count(count, **kwargs):
    state.double = 2 * int(count)


with DivLayout(server) as layout:
    html.Div("count = {{ count }}")
    html.Div("2 x count = {{ double }}")
    html.Input(
        type="range",
        min=0,
        max=10,
        step=1,
        v_model="count",
    )

server.start()



# Deep Reactive does not work with vue2

from trame.app import get_server
from trame.widgets import html, client
from trame.ui.html import DivLayout

server = get_server()
server.client_type = "vue2"
state, ctrl = server.state, server.controller

state.nested_1 = {
    "sliders": [
        {"value": 10, "min": 1, "max": 20, "step": 2},
        {"value": 1, "min": 1, "max": 20, "step": 1},
        {"value": 5, "min": 1, "max": 20, "step": 5},
    ],
    "a": 10,
}
state.nested_2 = {
    "sliders": [
        {"value": 10, "min": 1, "max": 20, "step": 2},
        {"value": 1, "min": 1, "max": 20, "step": 1},
        {"value": 5, "min": 1, "max": 20, "step": 5},
    ],
    "a": 10,
}

state.nested_3 = {
    "sliders": [
        {"value": 10, "min": 1, "max": 20, "step": 2},
        {"value": 1, "min": 1, "max": 20, "step": 1},
        {"value": 5, "min": 1, "max": 20, "step": 5},
    ],
    "a": 10,
}


@state.change("nested_1")
def update_nested_1(nested_1, **kwargs):
    print("nested 1")


@state.change("nested_2")
def update_nested_2(nested_2, **kwargs):
    print("nested 2")


@state.change("nested_3")
def update_nested_3(nested_3, **kwargs):
    print("nested 3")


with DivLayout(server) as layout:
    # Workaround / old fashion
    html.Input(
        type="range",
        value=("nested_2.a",),
        raw_attrs=[
            "@input=\"nested_2.a = $event.target.value; flushState('nested_2')\""
        ],
        min=1,
        max=10,
        step=1,
    )
    with html.Div("Nested 1 - {{ nested_1.a }}"):
        html.Input(
            type="range",
            value=("nested_1.a",),
            raw_attrs=[
                "@input=\"nested_1.a = $event.target.value; flushState('nested_1')\""
            ],
            min=1,
            max=10,
            step=1,
        )
        with html.Div("Sliders"):
            html.Input(
                v_for="s, i in nested_1.sliders",
                key="i",
                type="range",
                value=("s.value",),
                min=("s.min",),
                max=("s.max",),
                step=("s.step",),
                raw_attrs=[
                    "@input=\"s.value = $event.target.value; flushState('nested_1')\""
                ],
            )

    # New deep reactive
    with client.DeepReactive("nested_2") as dr:
        with html.Div("Nested 2 - {{ nested_2.a }}"):
            html.Input(type="range", v_model="nested_2.a", min=1, max=10, step=1)
            with html.Div("Sliders"):
                html.Input(
                    v_for="s, i in nested_2.sliders",
                    key="i",
                    type="range",
                    v_model="s.value",
                    min=("s.min",),
                    max=("s.max",),
                    step=("s.step",),
                )

    # Not working
    with html.Div("Nested 3 - {{ nested_3.a }}"):
        html.Input(type="range", v_model="nested_3.a", min=1, max=10, step=1)
        with html.Div("Sliders"):
            html.Input(
                v_for="s, i in nested_3.sliders",
                key="i",
                type="range",
                v_model="s.value",
                min=("s.min",),
                max=("s.max",),
                step=("s.step",),
            )

server.start()



import json
import sys
from pathlib import Path
from xprocess import ProcessStarter
from PIL import Image
from pixelmatch.contrib.PIL import pixelmatch

from playwright.sync_api import expect, Page


# ---------------------------------------------------------
# Pytest helpers
# ---------------------------------------------------------


class TrameServerMonitor:
    def __init__(self, log_path, do_print_log_lines=False):
        self._log_path = log_path
        self._last_state = {}
        self.port = 0
        self.do_print_log_lines = do_print_log_lines
        self.update()

    def update(self):
        last_state_line = "STATE: {}"
        with open(self._log_path, "r") as f:
            for line in f.readlines():
                if self.do_print_log_lines:
                    print(line)
                if "SERVER_PORT:" in line:
                    self.port = int(line[13:])
                if line[:7] == "STATE: ":
                    last_state_line = line

        self._last_state = json.loads(last_state_line[7:])

    def get_state(self):
        self.update()
        return self._last_state

    def get(self, name):
        self.update()
        return self._last_state.get(name)


def print_state(**kwargs):
    print("STATE:", json.dumps(kwargs), flush=True)


def enable_testing(server, *state_monitor):
    server.state.change(*state_monitor)(print_state)

    @server.controller.add("on_server_ready")
    def print_server_port(**kwargs):
        print("SERVER_PORT:", server.port, flush=True)

    return server


class FixtureHelper:
    def __init__(self, root_path):
        self.root_path = Path(root_path)

    def remove_page_urls(self):
        BASE_PATH = self.root_path / "visual_baseline"
        for file in BASE_PATH.glob("**/page_url.txt"):
            file.unlink()
            print(f" - remove: {file}")

    def get_xprocess_args(self, server_path):
        class Starter(ProcessStarter):
            terminate_on_interrupt = True
            pattern = "App running at:"

            # command to start process
            args = [
                Path(sys.executable).as_posix(),
                str(self.root_path / server_path),
                "--server",
                "--host",
                "127.0.0.1",
                "--port",
                "0",
            ]

        return Path(server_path).name, Starter, TrameServerMonitor


def assert_images_match(img_test: Image, ref_path: Path, threshold=0.1):
    img_ref = Image.open(ref_path)
    img_diff = Image.new("RGBA", img_ref.size)
    diff_path = ref_path.parent / f"diff_{ref_path.with_suffix('.png').name}"

    mismatch = pixelmatch(img_ref, img_test, img_diff, threshold)
    img_diff.save(diff_path)
    H, W = img_ref.size
    assert mismatch <= threshold * H * W


# ---------------------------------------------------------
# Playwright helpers
# ---------------------------------------------------------


def assert_screenshot_matches(
    page: Page, ref_dir: Path, name: str, threshold: float = 0.1
):
    img_dir = ref_dir / name
    ref_images = list(img_dir.glob("ref*.png"))
    if not ref_images:
        print(f"No reference images exist in {img_dir}. Creating one...")
        img_dir.mkdir(parents=True, exist_ok=True)
        ref_img_path = img_dir / "ref1.png"
        page.screenshot(path=ref_img_path)
        return

    # Save the test image
    test_img_path = img_dir / "test_image.png"
    page.screenshot(path=test_img_path)
    img_test = Image.open(test_img_path)

    # If there are reference images, find one that matches
    # Only write out image diffs if none succeed
    for ref_img_path in ref_images:
        try:
            # `expect(page).to_have_screenshot()` is not yet supported
            # in Python playwright, even though it is available in
            # JavaScript... :-\
            # Use our own comparison for now.
            # expect(page).to_have_screenshot(threshold=threshold)
            assert_images_match(img_test, ref_img_path, threshold)
        except AssertionError:
            # We'll just try the next one
            continue
        else:
            # It matched. Return!
            return

    raise AssertionError(f"No reference images matched in {img_dir}")


def assert_snapshot_matches(page: Page, ref_dir: Path, name: str):
    html = page.locator("html")
    ref_path = ref_dir / f"{name}.yml"
    if not ref_path.exists():
        print(f"'{ref_path}' does not exist. Creating...")
        with open(ref_path, "w") as wf:
            wf.write(html.aria_snapshot())
        return

    with open(ref_path, "r") as rf:
        expect(html).to_match_aria_snapshot(rf.read())



from trame.app import get_server
from trame_client.ui.html import DivLayout
from trame_client.widgets import html

from trame_server.utils.namespace import Translator

CLIENT_TYPE = "vue3"


class BasicApp:
    def __init__(self, server=None, template_name=None, client_type=CLIENT_TYPE):
        self.server = get_server(server, client_type=client_type)
        self._ui = None
        self.template_name = template_name
        self.state.a = 1
        self.state["b"] = 2
        self.state.update(
            {
                "c": 3,
                "d": 4,
                "longer_name": "hello",
            }
        )
        self.state.setdefault("e", 5)

        self.ctrl.plus = self.add
        self.ctrl.minus = self.remove

    def add(self):
        self.state.f += 1

    def remove(self):
        self.state.f -= 1

    def ui_content(self):
        html.Input(type="range", v_model_number=("f", 10), min=-10, max=10, step=1)
        html.Div(
            "Bigger than 5 {{ longer_name }}",
            v_show="Math.abs(f)>5 && longer_name.length",
        )
        html.Div(
            "Current value {{ Math.abs(f) + f }} <=> {{ f }} | {{ longer_name }}",
        )

    @property
    def ui(self):
        if self._ui is None:
            with DivLayout(self.server, template_name=self.template_name) as layout:
                self._ui = layout
                self.ui_content()

        return self._ui

    @property
    def nestable_ui(self):
        if self._ui is None:
            with html.Div(trame_server=self.server) as div:
                self._ui = div
                self.ui_content()

        return self._ui

    @property
    def state(self):
        return self.server.state

    @property
    def ctrl(self):
        return self.server.controller


class NestedApp:
    def __init__(self, server=None, client_type=CLIENT_TYPE):
        self.server = get_server(server, client_type=client_type)
        self._ui = None
        self.child_a_app = BasicApp(
            self.server.create_child_server(prefix="child_a_"), client_type=client_type
        )
        self.child_b_app = BasicApp(
            self.server.create_child_server(prefix="child_b_"), client_type=client_type
        )

    @property
    def ui(self):
        if self._ui is None:
            with DivLayout(self.server) as layout:
                self._ui = layout
                self.child_a_app.nestable_ui
                self.child_b_app.nestable_ui

        return self._ui

    @property
    def state(self):
        return self.server.state

    @property
    def ctrl(self):
        return self.server.controller


def test_prefix():
    root_server = get_server("test_prefix", client_type=CLIENT_TYPE)
    main_app = BasicApp(root_server)
    child_a_app = BasicApp(root_server.create_child_server(prefix="child_a_"))
    child_b_app = BasicApp(root_server.create_child_server(prefix="child_b_"))

    main_app.state.ready()
    child_a_app.state.ready()
    child_b_app.state.ready()

    main_app.ui
    child_a_app.ui
    child_b_app.ui

    global_state = {
        k: v for k, v in main_app.state.to_dict().items() if not k.startswith("trame_")
    }
    expected = dict(
        a=1,
        b=2,
        c=3,
        d=4,
        e=5,
        f=10,
        longer_name="hello",
        child_a_a=1,
        child_a_b=2,
        child_a_c=3,
        child_a_d=4,
        child_a_e=5,
        child_a_f=10,
        child_a_longer_name="hello",
        child_b_a=1,
        child_b_b=2,
        child_b_c=3,
        child_b_d=4,
        child_b_e=5,
        child_b_f=10,
        child_b_longer_name="hello",
    )
    assert expected == global_state

    main_app.add()
    assert main_app.state.f == 11
    main_app.ctrl.plus()
    assert main_app.state.f == 12

    # with child_a_app.state:
    child_a_app.remove()
    assert child_a_app.state.f == 9
    child_a_app.remove()
    assert child_a_app.state.f == 8
    assert main_app.state.child_a_f == 8

    # with child_b_app.state:
    child_b_app.ctrl.minus()
    assert child_b_app.state.f == 9
    child_b_app.remove()
    assert child_b_app.state.f == 8
    assert main_app.state.child_b_f == 8

    # Test main app output
    expected_main = """<div >
<input v-model.number="f" max="10" min="-10" step="1" type="range" />
<div v-show="Math.abs(f)>5 && longer_name.length">
Bigger than 5 {{ longer_name }}
</div>
<div >
Current value {{ Math.abs(f) + f }} <=> {{ f }} | {{ longer_name }}
</div>
</div>"""
    assert main_app.ui.html == expected_main

    # Test child_a app output
    expected_child_a = """<div >
<input v-model.number="child_a_f" max="10" min="-10" step="1" type="range" />
<div v-show="Math.abs(child_a_f)>5 && child_a_longer_name.length">
Bigger than 5 {{ child_a_longer_name }}
</div>
<div >
Current value {{ Math.abs(child_a_f) + child_a_f }} <=> {{ child_a_f }} | {{ child_a_longer_name }}
</div>
</div>"""
    assert child_a_app.ui.html == expected_child_a

    # Test child_b app output
    expected_child_b = """<div >
<input v-model.number="child_b_f" max="10" min="-10" step="1" type="range" />
<div v-show="Math.abs(child_b_f)>5 && child_b_longer_name.length">
Bigger than 5 {{ child_b_longer_name }}
</div>
<div >
Current value {{ Math.abs(child_b_f) + child_b_f }} <=> {{ child_b_f }} | {{ child_b_longer_name }}
</div>
</div>"""
    assert child_b_app.ui.html == expected_child_b


def test_translation():
    root_server = get_server("test_translation", client_type=CLIENT_TYPE)
    main_app = BasicApp(root_server)

    a_translator = Translator()
    a_translator.add_translation("longer_name", "child_a_longer_name")
    child_a_app = BasicApp(root_server.create_child_server(translator=a_translator))

    b_translator = Translator()
    b_translator.add_translation("longer_name", "child_b_longer_name")
    child_b_app = BasicApp(root_server.create_child_server(translator=b_translator))

    main_app.state.ready()
    child_a_app.state.ready()
    child_b_app.state.ready()

    main_app.ui
    child_a_app.ui
    child_b_app.ui

    global_state = {
        k: v for k, v in main_app.state.to_dict().items() if not k.startswith("trame_")
    }
    expected = dict(
        a=1,
        b=2,
        c=3,
        d=4,
        e=5,
        f=10,
        longer_name="hello",
        child_a_longer_name="hello",
        child_b_longer_name="hello",
    )
    assert expected == global_state

    # f points to the same value in all 3 apps
    assert main_app.state.f == 10
    assert child_a_app.state.f == 10
    assert child_a_app.state.f == 10
    main_app.add()
    assert main_app.state.f == 11
    assert child_a_app.state.f == 11
    assert child_a_app.state.f == 11
    main_app.ctrl.plus()
    assert main_app.state.f == 12
    assert child_a_app.state.f == 12
    assert child_a_app.state.f == 12

    # with child_a_app.state:
    child_a_app.remove()
    assert main_app.state.f == 11
    assert child_a_app.state.f == 11
    assert child_a_app.state.f == 11
    child_a_app.ctrl.minus()
    assert main_app.state.f == 10
    assert child_a_app.state.f == 10
    assert child_a_app.state.f == 10

    # with child_b_app.state:
    child_b_app.remove()
    assert main_app.state.f == 9
    assert child_a_app.state.f == 9
    assert child_a_app.state.f == 9
    child_b_app.ctrl.minus()
    assert main_app.state.f == 8
    assert child_a_app.state.f == 8
    assert child_a_app.state.f == 8

    # Test main app output
    expected_main = """<div >
<input v-model.number="f" max="10" min="-10" step="1" type="range" />
<div v-show="Math.abs(f)>5 && longer_name.length">
Bigger than 5 {{ longer_name }}
</div>
<div >
Current value {{ Math.abs(f) + f }} <=> {{ f }} | {{ longer_name }}
</div>
</div>"""
    assert main_app.ui.html == expected_main

    # Test child_a app output
    expected_child_a = """<div >
<input v-model.number="f" max="10" min="-10" step="1" type="range" />
<div v-show="Math.abs(f)>5 && child_a_longer_name.length">
Bigger than 5 {{ child_a_longer_name }}
</div>
<div >
Current value {{ Math.abs(f) + f }} <=> {{ f }} | {{ child_a_longer_name }}
</div>
</div>"""
    assert child_a_app.ui.html == expected_child_a

    # Test child_b app output
    expected_child_b = """<div >
<input v-model.number="f" max="10" min="-10" step="1" type="range" />
<div v-show="Math.abs(f)>5 && child_b_longer_name.length">
Bigger than 5 {{ child_b_longer_name }}
</div>
<div >
Current value {{ Math.abs(f) + f }} <=> {{ f }} | {{ child_b_longer_name }}
</div>
</div>"""
    assert child_b_app.ui.html == expected_child_b


def test_nested():
    root_server = get_server("test_nested", client_type=CLIENT_TYPE)
    main_app = NestedApp(root_server)

    main_app.state.ready()
    main_app.child_a_app.state.ready()
    main_app.child_b_app.state.ready()

    main_app.ui

    global_state = {
        k: v for k, v in main_app.state.to_dict().items() if not k.startswith("trame_")
    }
    expected = dict(
        child_a_a=1,
        child_a_b=2,
        child_a_c=3,
        child_a_d=4,
        child_a_e=5,
        child_a_f=10,
        child_a_longer_name="hello",
        child_b_a=1,
        child_b_b=2,
        child_b_c=3,
        child_b_d=4,
        child_b_e=5,
        child_b_f=10,
        child_b_longer_name="hello",
    )
    assert expected == global_state

    # Test main app output
    expected_main = """<div >
<div >
<input v-model.number="child_a_f" max="10" min="-10" step="1" type="range" />
<div v-show="Math.abs(child_a_f)>5 && child_a_longer_name.length">
Bigger than 5 {{ child_a_longer_name }}
</div>
<div >
Current value {{ Math.abs(child_a_f) + child_a_f }} <=> {{ child_a_f }} | {{ child_a_longer_name }}
</div>
</div>
<div >
<input v-model.number="child_b_f" max="10" min="-10" step="1" type="range" />
<div v-show="Math.abs(child_b_f)>5 && child_b_longer_name.length">
Bigger than 5 {{ child_b_longer_name }}
</div>
<div >
Current value {{ Math.abs(child_b_f) + child_b_f }} <=> {{ child_b_f }} | {{ child_b_longer_name }}
</div>
</div>
</div>"""
    assert main_app.ui.html == expected_main



import json
from pathlib import Path


def load_json(name):
    file_path = Path(__file__).parent.parent / Path(name)
    with open(file_path, "r") as f:
        return json.load(f)


def to_str(v):
    if isinstance(v, (tuple, list)):
        return v
    return f'"{v}"'


def to_doc_str(v):
    if isinstance(v, (tuple, list)):
        return v[0]
    return v


class HtmlEntry:
    def __init__(self, html_element, html_attributes, html_events):
        self.name = html_element.capitalize()
        self.html_element = html_element
        self.html_attributes = list(map(to_str, html_attributes))
        self.html_events = list(map(to_str, html_events))
        self._raw_props = html_attributes
        self._raw_events = html_events

    def __str__(self):
        lines = []
        lines.append(f"class {self.name}(HtmlElement):")
        lines.append('    """')
        lines.append("    Properties:\n")
        for name in self._raw_props:
            lines.append(f"    :param {to_doc_str(name)}:")
        lines.append("\n    Events:\n")
        for name in self._raw_events:
            lines.append(f"    :param {to_doc_str(name)}:")
        lines.append('\n    """')
        lines.append("    def __init__(self, children=None, **kwargs):")
        lines.append(
            f'        super().__init__("{self.html_element}", children, **kwargs)'
        )

        # List attributes
        if len(self.html_attributes):
            lines.append("        self._attr_names += [")
            for attr in self.html_attributes:
                lines.append(f"            {attr},")
            lines.append("        ]")

        # List events
        if len(self.html_attributes):
            lines.append("        self._event_names += [")
            for event in self.html_events:
                lines.append(f"            {event},")
            lines.append("        ]")

        lines.append("")
        return "\n".join(lines)


# -----------------------------------------------------------------------------
HTML_ATTRIBUTES = load_json("resources/attributes.json")
HTML_EVENTS = load_json("resources/events.json")
HTML_VUEJS = load_json("resources/vue.json")

HTML_ELEMENTS = set(HTML_ATTRIBUTES.keys())
HTML_ELEMENTS.discard("*")


def attributes(name):
    return (
        HTML_ATTRIBUTES.get("*", [])
        + HTML_VUEJS["attributes"].get("*", [])
        + HTML_ATTRIBUTES.get(name, [])
        + HTML_VUEJS["attributes"].get(name, [])
    )


def events(name):
    return (
        HTML_EVENTS.get("*", [])
        + HTML_VUEJS["events"].get("*", [])
        + HTML_EVENTS.get(name, [])
        + HTML_VUEJS["events"].get(name, [])
    )


def generate_html_elements():
    dst_file = Path(__file__).parent / Path("html.py")
    output = [
        "# Generated file - DO NOT EDIT",
        "from .core import (",
        "   HtmlElement,",
        "   Template,",
        "   Component,",
        "   Transition,",
        "   TransitionGroup,",
        "   KeepAlive,",
        "   Teleport,",
        "   Suspense,",
        ")",
        "from .. import module",
        "",
        "HtmlElement.MODULE = module",
        "",
    ]

    # Fill __all__
    output.append("__all__ = [")
    output.append('    "Template",')
    output.append('    "Component",')
    output.append('    "Transition",')
    output.append('    "TransitionGroup",')
    output.append('    "KeepAlive",')
    output.append('    "Teleport",')
    output.append('    "Suspense",')
    for elem in HTML_ELEMENTS:
        output.append(f'    "{elem.capitalize()}",')
    output.append("]\n\n")

    # Fill classes
    for elem in HTML_ELEMENTS:
        output.append(str(HtmlEntry(elem, attributes(elem), events(elem))))

    with open(dst_file, "w") as f:
        f.write("\n".join(output))


if __name__ == "__main__":
    print("Generating HTML elements...")
    generate_html_elements()
    print("Done...")



from trame.app import get_server
from trame.widgets import html
from trame.ui.html import DivLayout
from trame_client.utils.testing import enable_testing

server = get_server(client_type="vue3")
state, ctrl = server.state, server.controller

state.count = 1


@ctrl.set("update_ui")
def update_ui():
    state.count += 1
    with DivLayout(server):
        html.Div(f"Static text {state.count}", classes="staticDiv")
        html.Div("count = {{ count }}", classes="countDiv")
        html.Div("tts = {{ tts }}", classes="ttsDiv")
        html.Button("Update template", click=ctrl.update_ui, classes="updateBtn")
        html.Button("count++", click="count++", classes="plusBtn")


update_ui()
enable_testing(server, "count")
server.start()



from trame.app import get_server
from trame.widgets import client, html
from trame_client.ui.html import DivLayout
from trame_client.utils.testing import enable_testing

server = enable_testing(get_server(client_type="vue3"), "message")
state, ctrl = server.state, server.controller


def revert_message():
    state.message = state.message[::-1]


with DivLayout(server) as layout:
    eval_js = client.JSEval(
        event=("message", "hello world"),
        exec="window.document.querySelector('.jsAlert').innerHTML = $event",
    )
    ctrl.exec = eval_js.exec
    html.Div("{{ message }}")
    html.Div("Alert", classes="jsAlert")
    html.Button("Exec", click=ctrl.exec, classes="alertMsg")
    html.Button("Exec with arg", click=(ctrl.exec, "['Yes me']"), classes="alertMe")
    html.Button("Change message", click=revert_message, classes="swapMsg")

server.start()



from trame.app import get_server

import sys
import os

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "../../tests"))
from test_translator import NestedApp  # noqa: E402

CLIENT_TYPE = "vue2"
ISOLATED = False


def main():
    root_server = get_server(client_type=CLIENT_TYPE)

    root_app = NestedApp(root_server, client_type=CLIENT_TYPE)

    if not ISOLATED:
        root_app.child_a_app.server.translator.add_translation("f", "common_f")
        root_app.child_b_app.server.translator.add_translation("f", "common_f")

    root_app.ui

    root_server.controller.on_server_ready.add(
        lambda **kwargs: print("Root server ready")
    )
    root_app.server.start()


if __name__ == "__main__":
    main()



import sys
import logging

from trame_client.utils.defaults import TrameDefault
from trame_client.utils.formatter import to_pretty_html
from trame_common.obj.component import TrameComponent

AVAILABLE_DIRECTIVES = [
    ("v_text", "v-text"),
    ("v_html", "v-html"),
    ("v_show", "v-show"),
    ("v_if", "v-if"),
    ("v_else", "v-else"),
    ("v_else_if", "v-else-if"),
    ("v_for", "v-for"),
    ("v_on", "v-on"),
    ("v_bind", "v-bind"),
    ("v_model", "v-model"),
    ("v_model_lazy", "v-model.lazy"),
    ("v_model_number", "v-model.number"),
    ("v_model_trim", "v-model.trim"),
    ("v_slot", "v-slot"),
    ("v_pre", "v-pre"),
    ("v_once", "v-once"),
    ("v_memo", "v-memo"),
    ("v_cloak", "v-cloak"),
]
KEY_ALIAS = [
    "enter",
    "tab",
    "delete",
    "esc",
    "space",
    "left",
    "up",
    "right",
    "down",
]
KEY_MODIFIER = ["ctrl", "alt", "shift", "meta"]
MOUSE_BUTTONS = ["left", "right", "middle"]
V_ON_MODIFIER = [
    "stop",
    "prevent",
    "capture",
    "self",
    "once",
    "left",
    "right",
    "middle",
    "passive",
]  # *KEY_ALIAS
V_ON_TYPE_MOUSE = [
    "click",
    "contextmenu",
    "dblclick",
    "mousedown",
    "mouseenter",
    "mouseleave",
    "mousemove",
    "mouseup",
    "mouseout",
    "mouseover",
]
V_ON_TYPE_KEYBOARD = [
    "keydown",
    "keyup",
    "keypress",
]
V_ON_TYPE_FORM = [
    "submit",
    "input",
    "change",
    "focus",
    "blur",
]
V_ON_TYPE_TOUCH = [
    "touchstart",
    "touchmove",
    "touchend",
    "touchcancel",
]
V_ON_TYPE_UI = [
    "scroll",
    "resize",
    "select",
]
V_ON_TYPE_ANIM = [
    "animationstart",
    "animationend",
    "animationiteration",
]
V_ON_TYPE_TRANSITION = [
    "transitionstart",
    "transitionend",
    "transitioncancel",
]

V_MODEL_MODIFIER = {"lazy", "number", "trim"}

SHARED_ATTRIBUTES = [
    "accesskey",
    "autocapitalize",
    "autofocus",
    ("classes", "class"),
    "contenteditable",
    "dir",
    "draggable",
    "enterkeyhint",
    "hidden",
    "id",
    "inputmode",
    "is",
    "itemid",
    "itemprop",
    "itemref",
    "itemscope",
    "itemtype",
    "lang",
    "nonce",
    "slot",
    "spellcheck",
    "style",
    "tabindex",
    "title",
    "translate",
    # --
    "ref",
    ["key", ":key"],
]

# !all modifiers should go through v_on_click_left_stop...
SHARED_EVENTS = [
    *V_ON_TYPE_MOUSE,
    *V_ON_TYPE_KEYBOARD,
    *V_ON_TYPE_FORM,
    *V_ON_TYPE_TOUCH,
    *V_ON_TYPE_ANIM,
    *V_ON_TYPE_TRANSITION,
    # -------------------------------------------------------------------------
    # Should not enable by default as they tend to conflict with widget
    # properties. Use v_on_{}= to enable any missing mapping.
    # -------------------------------------------------------------------------
    # *V_ON_TYPE_UI,
]

logger = logging.getLogger(__name__)


def py2js_key(key):
    return key.replace("_", "-")


class ElementContextManager:
    def __init__(self):
        self.element_stack = []
        self._server = None

    def enter(self, elem):
        self._server = elem.server
        self.element_stack.append(elem)

    def exit(self, elem):
        if len(self.element_stack) and elem == self.element_stack[-1]:
            self.element_stack.pop()

        if len(self.element_stack):
            self._server = self.element_stack[-1].server

    def add_child(self, elem, connect_parent=True):
        if len(self.element_stack):
            if elem.server is None:
                elem.set_server(self.element_stack[-1].server)
            if connect_parent:
                self.element_stack[-1].add_child(elem)

        if elem.server is None:
            elem.set_server(self._server)

        if elem.server is None:
            print(f"No server available for element {elem._elem_name}")


HTML_CTX = ElementContextManager()


class VirtualNode:
    """
    A VirtualNode element allow to logically define UI element that
    can be filled or used in an independent order and yet update
    dynamically the layout upon change.
    """

    def __init__(self, trame_server=None, **_):
        self._server = trame_server
        self._children = []
        self._layouts = set()

    @property
    def server(self):
        """Return the associated server"""
        return self._server

    def set_server(self, v):
        """Update the associated server"""
        self._server = v

    def add_child(self, child):
        """
        Add a component to this component's children

        :param child: The component to add as a child
        :type child: str | AbstractElement
        """
        self._children.append(child)
        return self

    def add_children(self, children):
        """
        Add components to this component's children.
        The provided children is expected to be a list.

        :param children: The list of components to add to the children
        :type children: list
        """
        self._children += children
        return self

    @property
    def children(self):
        """
        Children components
        """
        return self._children

    def clear(self):
        """
        Remove all children
        """
        self._children.clear()
        return self

    @property
    def html(self):
        """
        Return a string representation of the HTML component
        """
        out_buffer = []
        try:
            for elem in self._children:
                if isinstance(elem, str):
                    out_buffer.append(elem)
                else:
                    out_buffer.append(elem.html)
        except Exception as e:
            print(e)

        return "\n".join(out_buffer)

    # -------------------------------------------------------------------------
    # Resource manager
    # -------------------------------------------------------------------------

    def __enter__(self):
        HTML_CTX.enter(self)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        HTML_CTX.exit(self)
        self.flush_content()

    # -------------------------------------------------------------------------
    # Layout handling
    # -------------------------------------------------------------------------

    def clear_layouts(self):
        """Remove any reference to previously registered layout"""
        self._layouts.clear()
        return self

    def flush_content(self):
        """Push VirtualNode content to registered layouts"""
        for layout in self._layouts:
            layout.flush_content()
        return self

    def __call__(self, layout=None, **kwargs):
        if layout is not None:
            self._layouts.add(layout)
            self._server = layout.server

        HTML_CTX.add_child(self)


def _event_value_processing(server, js_key, value):
    if isinstance(value, str):
        translated_value = server.state.translator.translate_js_expression(
            server.state, value
        )
        return f'{js_key}="{translated_value}"'
    elif callable(value):
        trigger_name = server.trigger_name(value)
        return f"{js_key}=\"trigger('{trigger_name}')\""
    elif isinstance(value, tuple):
        trigger_name = value[0]
        if callable(trigger_name):
            trigger_name = server.trigger_name(trigger_name)
        if len(value) == 1:
            return f"{js_key}=\"trigger('{trigger_name}')\""
        if len(value) == 2:
            translated_value = server.state.translator.translate_js_expression(
                server.state, value[1]
            )
            return f"{js_key}=\"trigger('{trigger_name}', {translated_value})\""
        if len(value) == 3:
            translated_value = server.state.translator.translate_js_expression(
                server.state, value[1]
            )
            # We don't want to translate kwargs as we may change keys rather than just values
            return f"{js_key}=\"trigger('{trigger_name}', {translated_value}, {value[2]})\""
    return False


class AbstractElement(TrameComponent):
    """
    A Vue component which can integrate with the rest of trame

    See Vue docs |vue_doc_link| for more info

    .. |vue_doc_link| raw:: html

        <a href="https://vuejs.org/v2/guide/instance.html" target="_blank">here</a>

    .. |mdn_doc_link| raw:: html

        <a href="https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes" target="_blank">here</a>

    .. |mdn_event_link| raw:: html

        <a href="https://developer.mozilla.org/en-US/docs/Web/API/Element#mouse_events">here</a>

    :param name: The name of the element, like 'div' for a ``<div/>`` element
    :type name: str
    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    :param __properties: Provide more attribute names that should be handle
    :param __events: Provide more event names that should be handle

    Html attributes - See |mdn_doc_link| for more info

    :param id: See |mdn_doc_link| for more info
    :param classes: Match the HTML `class` attribute. See |mdn_doc_link| for more info
    :param style: See |mdn_doc_link| for more info

    Vue attributes - See |vue_doc_link| for more info

    :param ref: See |vue_doc_link| for more info
    :param v_model: See |vue_doc_link| for more info
    :param v_if: See |vue_doc_link| for more info
    :param v_show: See |vue_doc_link| for more info
    :param v_for: See |vue_doc_link| for more info
    :param key: See |vue_doc_link| for more info

    Events - See |mdn_event_link| for more info

    :param click: See |mdn_event_link| for more info
    :param mousedown: See |mdn_event_link| for more info
    :param mouseup: See |mdn_event_link| for more info
    :param mouseenter: See |mdn_event_link| for more info
    :param mouseleave: See |mdn_event_link| for more info
    :param contextmenu: See |mdn_event_link| for more info

    Raw attributes

    :param raw_attrs: List of string that will be added as-is in the generated template

    >>> print(html.Template(raw_attrs=["v-slot:item.1", 'class="bg-red"', '@click.stop="a=2"']))
    ... <Template v-slot:item.1 class="bg-red" @click.stop="a=2" />

    Context Name:

    :param ctx_name: name to attach instance to server.context if provided

    DOM hierarchy handling:

    :param connect_parent: True by default which means any widget instantiation will connect
                           to the context manager widget that the current instance directly belong to.
                           In some specific case, it could be useful to disable that behavior by
                           setting it to False.
    """

    _next_id = 1
    _debug = "--debug" in sys.argv or "-d" in sys.argv

    def __init__(
        self,
        _elem_name,
        children=None,
        raw_attrs=None,
        ctx_name=None,
        connect_parent=True,
        **kwargs,
    ):
        AbstractElement._next_id += 1
        self._id = AbstractElement._next_id
        self._server = kwargs.get("trame_server")
        self._elem_name = _elem_name
        self._allowed_keys = set()
        self._attr_names = (
            kwargs.get("__properties", []) + SHARED_ATTRIBUTES + AVAILABLE_DIRECTIVES
        )
        self._event_names = kwargs.get("__events", []) + SHARED_EVENTS

        style = kwargs.get("style", None)
        if type(style) is dict:
            kwargs["style"] = " ".join([f"{k}: {v};" for k, v in style.items()])

        self._attributes = {}
        self._py_attr = kwargs
        self._children = []

        # Handle raw attributes if provided
        if raw_attrs:
            for idx, raw_value in enumerate(raw_attrs):
                self._attributes[f"_raw_{idx}"] = raw_value

        if children:
            if isinstance(children, (list, tuple)):
                self._children.extend(children)
            else:
                self._children.append(children)

        # Add ourself to context if any
        HTML_CTX.add_child(self, connect_parent)

        super().__init__(self._server, ctx_name=ctx_name)

    def _attr_str(self):
        return " ".join(self._attributes.values())

    def _update_allowed_keys(self):
        if hasattr(self, "_attr_names") and hasattr(self, "_event_names"):
            for items in [self._attr_names, self._event_names]:
                for item in items:
                    if isinstance(item, str):
                        self._allowed_keys.add(item)
                    else:
                        self._allowed_keys.add(item[0])

    # -------------------------------------------------------------------------
    # Static helper for widgets
    # -------------------------------------------------------------------------

    @staticmethod
    def register_directive(py_name, js_name=None):
        global AVAILABLE_DIRECTIVES
        if js_name is None:
            AVAILABLE_DIRECTIVES.append(py_name)
        else:
            AVAILABLE_DIRECTIVES.append((py_name, js_name))

    # -------------------------------------------------------------------------
    # App associated to HTML element
    # -------------------------------------------------------------------------

    def set_server(self, v):
        """Update the associated server"""
        self._server = v

    # -------------------------------------------------------------------------
    # Building API
    # -------------------------------------------------------------------------

    def __getitem__(self, name):
        return self._py_attr[name]

    def __setitem__(self, name, value):
        if name in self._allowed_keys:
            if value is None:
                self._py_attr.pop(name, value)
                self._attributes.pop(name, value)
            else:
                self._py_attr[name] = value

        else:
            print(f"Attribute {name} is not defined for {self._elem_name}")

    def __getattr__(self, name):
        if name[0] == "_":
            raise AttributeError()

        if name in self._py_attr:
            return self._py_attr[name]

        print(f"{self._elem_name}.{name} is missing")

    def __setattr__(self, name, value):
        if name[0] == "_":
            self.__dict__[name] = value
        elif name == "children":
            self._children = value
        elif name in self._allowed_keys:
            if value is None:
                self._py_attr.pop(name, value)
                self._attributes.pop(name, value)
            else:
                self._py_attr[name] = value

        else:
            super().__setattr__(name, value)

        if name in ["_attr_names", "_event_names"]:
            self._update_allowed_keys()

    # -------------------------------------------------------------------------
    # helpers
    # -------------------------------------------------------------------------

    def ttsSensitive(self):
        """
        Calling this function on an element will make it fully recreate itself
        every time the layout update. Internally it is managed by adding a `key=`
        attribute which use a layout timestamp.

        This is especially useful for component that manage other elements outside
        of themself like VSelect in Vuetify.
        """
        self._attributes["__tts"] = f':key="`w{self._id}-${{tts}}`"'
        return self

    def attrs(self, *names):
        """
        Calling this function will process the provided attribute names and
        configure its internal so the matching HTML string could easily be
        generated later on.

        :param names: The names attribute to process
        :type names: *str
        """
        directives = [
            name
            for name in self._py_attr.keys()
            if name.startswith("v_model_") or name.startswith("v_bind_")
        ]
        for _name in [*directives, *names]:
            js_key = None
            name = _name
            if isinstance(_name, (list, tuple)):
                name = _name[0]
                js_key = _name[1]

            if name in self._py_attr:
                if js_key is None:
                    js_key = py2js_key(name)
                value = self._py_attr[name]

                # smart key handling
                if name.startswith("v_model_"):
                    model_name, *modifiers = name.split("_")[2:]
                    if model_name in V_MODEL_MODIFIER and len(modifiers) == 0:
                        js_key = f"v-model.{model_name}"
                    else:
                        js_key = f"v-model:{model_name}{'.' if len(modifiers) else ''}{'.'.join(modifiers)}"
                elif name.startswith("v_bind_"):
                    prop_name, *modifiers = name.split("_")[2:]
                    js_key = f":{prop_name}{'.' if len(modifiers) else ''}{'.'.join(modifiers)}"

                if value is None:
                    continue

                logger.info("js_key = %s", js_key)

                if isinstance(value, (tuple, list)):
                    if len(value) > 1:
                        if isinstance(value[1], TrameDefault):
                            value[1].set_defaults(self.server)
                        else:
                            self.server.state.setdefault(value[0], value[1])

                    logger.info("before: %s = %s", js_key, value[0])
                    if isinstance(value[0], str):
                        translated_value = (
                            self.server.state.translator.translate_js_expression(
                                self.server.state, value[0]
                            )
                        )
                    else:
                        translated_value = str(value[0])
                        if AbstractElement._debug:
                            logger.warning(
                                'Warning: <%s %s="..." /> is set with an (%s)',
                                self._elem_name,
                                js_key,
                                type(value[0]),
                            )

                    logger.info("after: %s = %s", js_key, translated_value)
                    if js_key.startswith("v-"):
                        self._attributes[name] = f'{js_key}="{translated_value}"'
                    elif js_key.startswith(":"):
                        self._attributes[name] = f'{js_key}="{translated_value}"'
                    else:
                        self._attributes[name] = f':{js_key}="{translated_value}"'
                elif isinstance(value, bool):
                    if value:
                        self._attributes[name] = js_key
                    else:
                        self._attributes[name] = f':{js_key}="false"'
                elif isinstance(value, str):
                    if js_key.startswith("v-") or js_key.startswith(":"):
                        logger.info("before: %s = %s", js_key, value)
                        value = self.server.state.translator.translate_js_expression(
                            self.server.state, value
                        )
                        logger.info("after: %s = %s", js_key, value)

                    self._attributes[name] = f'{js_key}="{value}"'
                elif isinstance(value, (int, float)):
                    self._attributes[name] = f'{js_key}="{value}"'
                else:
                    print(
                        "Error: Don't know how to handle attribute name "
                        f"'{name}' with value '{value}' in {self.__class__}::{self._elem_name}"
                    )

        return self

    def events(self, *names):
        """
        Calling this function will process the provided event names and
        configure its internal so the matching HTML string could easily be
        generated later on.

        :param names: The names events to process
        :type names: *str
        """
        processed_event = set()
        for _name in names:
            js_key = None
            name = _name
            if isinstance(_name, tuple):
                name = _name[0]
                js_key = _name[1]
            if name in self._py_attr:
                if js_key is None:
                    js_key = py2js_key(name)
                js_key = f"@{js_key}"
                value = self._py_attr[name]

                if value is None:
                    continue

                attribute = _event_value_processing(self.server, js_key, value)
                if attribute is None:
                    # no match
                    pass
                elif isinstance(attribute, str):
                    self._attributes[name] = attribute
                    processed_event.add(name)
                else:
                    print(
                        "Error: Don't know how to handle event name "
                        f"'{name}' with value '{value}' in {self.__class__}::{self._elem_name}"
                    )

        # process v_on_....
        for key_name in self._py_attr:
            if key_name in processed_event:
                continue

            if key_name.startswith("v_on_"):
                tokens = key_name.split("_")[2:]
                js_key = f"@{'.'.join(tokens)}"
                value = self._py_attr[key_name]

                if value is None:
                    continue

                attribute = _event_value_processing(self.server, js_key, value)
                if attribute is None:
                    # no match
                    pass
                elif isinstance(attribute, str):
                    self._attributes[key_name] = attribute
                else:
                    print(
                        "Error: Don't know how to handle event name "
                        f"'{key_name}' with value '{value}' in {self.__class__}::{self._elem_name}"
                    )

        return self

    def clear(self):
        """
        Remove all children
        """
        self._children.clear()
        return self

    def hide(self):
        """
        Hide element while keeping it in the DOM. (display: none)
        """
        self._attributes["__style"] = 'style="display: none"'

    def add_child(self, child):
        """
        Add a component to this component's children

        :param child: The component to add as a child
        :type child: str | AbstractElement
        """
        self._children.append(child)

    def add_children(self, children):
        """
        Add components to this component's children.
        The provided children is expected to be a list.

        :param children: The list of components to add to the children
        :type children: list
        """
        self._children += children

    @property
    def children(self):
        """
        Children components
        """
        return self._children

    def set_text(self, value):
        """
        Replace children with a single text child element

        :param value: The text for the new text child element
        :type value: str
        """
        self.clear()
        self._children.append(value)

    @property
    def html(self):
        """
        Return a string representation of the HTML component
        """
        try:
            # Build attributes
            self.attrs(*self._attr_names)
            self.events(*self._event_names)

            # Patch ref to use trame registration
            if (
                self.server.client_type == "vue3"
                and "ref" in self._attributes
                and self._attributes["ref"].startswith("ref=")
            ):
                ref_name = self._attributes["ref"][5:-1]
                self._attributes["ref"] = (
                    f''':ref="(el) => trame.refs['{ref_name}'] = el"'''
                )

            # Compute HTML str
            if len(self._children):
                out_buffer = []
                out_buffer.append(f"<{self._elem_name} {self._attr_str()}>")
                for child in self._children:
                    if isinstance(child, str):
                        translated_value = (
                            self.server.state.translator.translate_vue_templating(
                                self.server.state, child
                            )
                        )
                        out_buffer.append(translated_value)
                    else:
                        out_buffer.append(child.html)
                out_buffer.append(f"</{self._elem_name}>")
                return "\n".join(out_buffer)
            else:
                return f"<{self._elem_name} {self._attr_str()} />"
        except Exception as e:
            logger.error(e)
            return f"<{self._elem_name} html-error />"

    def __repr__(self):
        return to_pretty_html(self.html)

    # -------------------------------------------------------------------------
    # Resource manager
    # -------------------------------------------------------------------------

    def __enter__(self):
        HTML_CTX.enter(self)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        HTML_CTX.exit(self)


class HtmlElement(AbstractElement):
    MODULE = None

    def __init__(self, _elem_name, children=None, **kwargs):
        super().__init__(_elem_name, children, **kwargs)
        if HtmlElement.MODULE and self.server:
            self.server.enable_module(HtmlElement.MODULE)


# -----------------------------------------------------------------------------
# Vue based components
# -----------------------------------------------------------------------------


class Template(AbstractElement):
    """
    The standard html content template element. This is mostly used by |slot_doc_link|.

    .. |slot_doc_link| raw:: html

        <a href="https://vuejs.org/v2/guide/instance.html" target="_blank">vue's slot system</a>


    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    :param v_slot: The slot this template corresponds to
    """

    slot_names = set()

    def __init__(self, children=None, **kwargs):
        super().__init__("template", children, **kwargs)
        self._attr_names += ["v_slot"]
        for slot_name in Template.slot_names:
            safe_name = slot_name.replace("-", "_").replace(".", "_")
            if "<name>" in safe_name:
                safe_header, safe_tail = safe_name.split("<name>")
                header, tail = slot_name.split("<name>")
                for key in kwargs:
                    if key.startswith(header):
                        dyna_name = key[len(header) : -len(tail)]
                        self._attr_names.append(
                            (
                                f"v_slot_{safe_header}{dyna_name}{safe_tail}",
                                f"v-slot:{header}{dyna_name}{tail}",
                            )
                        )
            else:
                self._attr_names.append((f"v_slot_{safe_name}", f"v-slot:{slot_name}"))


# -----------------------------------------------------------------------------


class Component(AbstractElement):
    """
    Vue.js dynamic component element.

    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    :param is_name: name of the component to map
    """

    def __init__(self, children=None, **kwargs):
        super().__init__("component", children, **kwargs)
        self._attr_names += [
            ("is_name", "is"),
        ]


# -----------------------------------------------------------------------------


class Transition(AbstractElement):
    """
    Vue.js Transition component element.
    Provides animated transition effects to a single element or component.

    Properties:

    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    :param name: Used to automatically generate transition CSS class names.
        e.g. `name: 'fade'` will auto expand to `.fade-enter`,
        `.fade-enter-active`, etc.
    :param css: Whether to apply CSS transition classes. (Default: true)
    :param type: Specifies the type of transition events to wait for to
        determine transition end timing. Default behavior is auto detecting
        the type that has longer duration. (transition, animation)
    :param duraction: Specifies explicit durations of the transition.
        Default behavior is wait for the first `transitionend` or
        `animationend` event on the root transition element.
    :param mode: Controls the timing sequence of leaving/entering transitions.
        Default behavior is simultaneous. ('in-out' | 'out-in' | 'default')
    :param appear: Whether to apply transition on initial render.
        (default: false)
    :param enter_from_class: Props for customizing transition classes.
    :param enter_active_class: Props for customizing transition classes.
    :param enter_to_class: Props for customizing transition classes.
    :param appear_from_class: Props for customizing transition classes.
    :param appear_active_class: Props for customizing transition classes.
    :param appear_to_class: Props for customizing transition classes.
    :param leave_from_class: Props for customizing transition classes.
    :param leave_active_class: Props for customizing transition classes.
    :param leave_to_class: Props for customizing transition classes.

    Events:

    :param before_enter:
    :param before_leave:
    :param enter:
    :param leave:
    :param appear:
    :param after_enter:
    :param after_leave:
    :param after_appear:
    :param enter_cancelled:
    :param leave_cancelled:
    :param appear_cancelled:

    """

    def __init__(self, children=None, **kwargs):
        super().__init__("transition", children, **kwargs)
        self._attr_names += [
            "name",
            "css",
            "type",
            "duraction",
            "mode",
            "appear",
            ("enter_from_class", "enterFromClass"),
            ("enter_active_class", "enterActiveClass"),
            ("enter_to_class", "enterToClass"),
            ("appear_from_class", "appearFromClass"),
            ("appear_active_class", "appearActiveClass"),
            ("appear_to_class", "appearToClass"),
            ("leave_from_class", "leaveFromClass"),
            ("leave_active_class", "leaveActiveClass"),
            ("leave_to_class", "leaveToClass"),
        ]
        self._event_names += [
            ("before_enter", "before-enter"),
            ("before_leave", "before-leave"),
            ("enter", "enter"),
            ("leave", "leave"),
            ("appear", "appear"),
            ("after_enter", "after-enter"),
            ("after_leave", "after-leave"),
            ("after_appear", "after-appear"),
            ("enter_cancelled", "enter-cancelled"),
            ("leave_cancelled", "leave-cancelled"),
            ("appear_cancelled", "appear-cancelled"),
        ]


# -----------------------------------------------------------------------------


class TransitionGroup(AbstractElement):
    """
    Vue.js TransitionGroup component element.
    Provides transition effects for multiple elements or components in a list.

    Properties:

    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None

    Events:

    :param before_enter:
    :param before_leave:
    :param enter:
    :param leave:
    :param appear:
    :param after_enter:
    :param after_leave:
    :param after_appear:
    :param enter_cancelled:
    :param leave_cancelled:
    :param appear_cancelled:

    """

    def __init__(self, children=None, **kwargs):
        super().__init__("transition-group", children, **kwargs)
        self._attr_names += [
            "tag",
            ("move_class", "moveClass"),
        ]
        self._event_names += [
            ("before_enter", "before-enter"),
            ("before_leave", "before-leave"),
            ("enter", "enter"),
            ("leave", "leave"),
            ("appear", "appear"),
            ("after_enter", "after-enter"),
            ("after_leave", "after-leave"),
            ("after_appear", "after-appear"),
            ("enter_cancelled", "enter-cancelled"),
            ("leave_cancelled", "leave-cancelled"),
            ("appear_cancelled", "appear-cancelled"),
        ]


# -----------------------------------------------------------------------------


class KeepAlive(AbstractElement):
    """
    Vue.js KeepAlive component element.
    Caches dynamically toggled components wrapped inside.

    Properties:

    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    :param include: If specified, only components with names matched
        by `include` will be cached.
    :param exclude: Any component with a name matched by `exclude`
        will not be cached.
    :param max: The maximum number of component instances to cache.

    """

    def __init__(self, children=None, **kwargs):
        super().__init__("keep-alive", children, **kwargs)
        self._attr_names += [
            "include",
            "exclude",
            "max",
        ]


# -----------------------------------------------------------------------------


class Teleport(AbstractElement):
    """
    Vue.js Teleport component element.
    Renders its slot content to another part of the DOM.

    Properties:

    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    :param to: Required. Specify target container.
        Can either be a selector or an actual element.
    :param disabled: When `true`, the content will remain in its original
        location instead of moved into the target container.
        Can be changed dynamically.
    :param defer: When `true`, the Teleport will defer until other parts
        of the application have been mounted before resolving its target.

    """

    def __init__(self, children=None, **kwargs):
        super().__init__("teleport", children, **kwargs)
        self._attr_names += [
            "to",
            "disabled",
            "defer",
        ]


# -----------------------------------------------------------------------------


class Suspense(AbstractElement):
    """
    Vue.js Suspense component element.
    Used for orchestrating nested async dependencies in a component tree.

    Properties:

    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None
    :type timeout:
    :type suspensible:

    Events:

    :param resolve:
    :param pending:
    :param fallback:

    """

    def __init__(self, children=None, **kwargs):
        super().__init__("suspense", children, **kwargs)
        self._attr_names += [
            "timeout",
            "suspensible",
        ]
        self._event_names += [
            "resolve",
            "pending",
            "fallback",
        ]



from trame.app import get_server

import sys
import os

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "../../tests"))
from test_translator import BasicApp  # noqa: E402

CLIENT_TYPE = "vue2"


def main():
    root_server = get_server(client_type=CLIENT_TYPE)
    one_of_the_app = None
    for app_name in ["main", "a", "b", "c"]:
        child_server = root_server.create_child_server(prefix=f"{app_name}_")
        app = BasicApp(child_server, app_name, CLIENT_TYPE)
        app.ui
        app.add()
        app.ctrl.plus()
        one_of_the_app = app

    # root_server.controller.on_server_ready.add(
    #   lambda **kwargs: print(json.dumps(root_server.state.to_dict(), indent=2))
    # )
    root_server.controller.on_server_ready.add(
        lambda **kwargs: print("Root server ready")
    )
    one_of_the_app.server.controller.on_server_ready.add(
        lambda **kwargs: print("Sub server ready")
    )
    one_of_the_app.server.start()


if __name__ == "__main__":
    main()



from trame.app import get_server
from trame.widgets import client, vuetify
from trame.ui.vuetify import SinglePageLayout

# -----------------------------------------------------------------------------
# Trame setup
# -----------------------------------------------------------------------------

server = get_server(client_type="vue2")
state, ctrl = server.state, server.controller

state.message_1 = "Hello 1"
state.message_2 = "Hello 2"
state.message_3 = "Hello 3"

state.list_1 = [True, False, False]
state.list_2 = [False, True, False]
state.list_3 = [False, False, True]

state.msg_idx = 1


# -----------------------------------------------------------------------------
# UI setup
# -----------------------------------------------------------------------------

with SinglePageLayout(server) as layout:
    # Toolbar
    with layout.toolbar as toolbar:
        vuetify.VSpacer()
        vuetify.VBtn("1", click="msg_idx = 1")
        vuetify.VBtn("2", click="msg_idx = 2")
        vuetify.VBtn("3", click="msg_idx = 3")
        vuetify.VBtn(
            "Revert",
            click="""
                const name = `message_${msg_idx}`;
                set(name, get(name).split('').reverse().join(''))
            """,
        )

    with layout.content:
        with vuetify.VContainer():
            with vuetify.VRow():
                with vuetify.VCol():
                    with vuetify.VCard(style="height: 200px;"):
                        vuetify.VCardTitle("get function")
                        vuetify.VDivider()
                        vuetify.VCardText(">>> {{ get(`message_${msg_idx}`) }}")

                with vuetify.VCol():
                    with client.Getter(
                        name=("`message_${msg_idx}`",), key_name="var_key"
                    ):
                        with vuetify.VCard(style="height: 200px;"):
                            vuetify.VCardTitle("getter {{ var_key }}")
                            vuetify.VDivider()
                            vuetify.VCardText(">>> {{ value }}")

                with vuetify.VCol():
                    with vuetify.VCard(style="height: 200px;"):
                        vuetify.VCardTitle("getter with rename")
                        vuetify.VDivider()
                        with client.Getter(
                            name=("`message_${msg_idx}`",),
                            value_name="msg",
                            key_name="name",
                        ):
                            vuetify.VCardText(">>> {{ name }}={{ msg }}")

            with vuetify.VRow():
                with vuetify.VCol():
                    with client.Getter(
                        name=("`list_${msg_idx}`",),
                        key_name="name",
                        update_nested_name="update",
                    ):
                        with vuetify.VCard():
                            vuetify.VCardTitle("List {{ name }}")
                            vuetify.VDivider()
                            with vuetify.VCardText():
                                with vuetify.VList():
                                    with vuetify.VListItem(
                                        v_for="v, i in value", key="i"
                                    ):
                                        vuetify.VIcon(
                                            v_text="v ? 'mdi-eye' : 'mdi-eye-off'",
                                            click="update(i, !v)",
                                        )


# -----------------------------------------------------------------------------
# start server
# -----------------------------------------------------------------------------

if __name__ == "__main__":
    server.start()
